# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LP Solver for two-player zero-sum games."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import cvxopt
import numpy as np
from open_spiel.python.egt import utils
import pyspiel

# Constants that determine the type of objective (max vs. min) and type of
# constraints (<=, >=, =).
OBJ_MAX = 1
OBJ_MIN = 2
CONS_TYPE_LEQ = 3
CONS_TYPE_GEQ = 4
CONS_TYPE_EQ = 5

# Constants that determine the type of dominance to find.
DOMINANCE_STRICT = 1
DOMINANCE_VERY_WEAK = 2
DOMINANCE_WEAK = 3


class _Variable(object):
  """A variable in an LP."""

  def __init__(self, vid, lb=None, ub=None):
    """Creates a variable in a linear program.

    Args:
      vid: (integer) the variable id (should be unique for each variable)
      lb: the lower bound on the variable's value (None means no lower bound)
      ub: the upper bound on the variable's valie (None means no upper bound)
    """
    self.vid = vid
    self.lb = lb
    self.ub = ub


class _Constraint(object):
  """A constraint in an LP."""

  def __init__(self, cid, ctype):
    """Creates a constraint in a linear program.

    Args:
      cid: (integer) the constraint id (should be unique for each constraint)
      ctype: the constraint type (CONS_TYPE_{LEQ, GEQ, EQ})
    """
    self.cid = cid
    self.ctype = ctype
    self.coeffs = {}  # var label -> value
    self.rhs = None


class LinearProgram(object):
  """A object used to provide a user-friendly API for building LPs."""

  def __init__(self, objective):
    assert objective == OBJ_MIN or objective == OBJ_MAX
    self._valid_constraint_types = [CONS_TYPE_EQ, CONS_TYPE_LEQ, CONS_TYPE_GEQ]
    self._objective = objective
    self._obj_coeffs = {}  # var label -> value
    self._vars = {}  # var label -> var
    self._cons = {}  # cons label -> constraint
    self._var_list = []
    self._leq_cons_list = []
    self._eq_cons_list = []

  def add_or_reuse_variable(self, label, lb=None, ub=None):
    """Adds a variable to this LP, or reuses one if the label exists.

    If the variable already exists, simply checks that the upper and lower
    bounds are the same as previously specified.

    Args:
      label: a label to assign to this constraint
      lb: a lower-bound value for this variable
      ub: an upper-bound value for this variable
    """
    var = self._vars.get(label)
    if var is not None:
      # Do not re-add, but ensure it's the same
      assert var.lb == lb and var.ub == ub
      return
    var = _Variable(len(self._var_list), lb, ub)
    self._vars[label] = var
    self._var_list.append(var)

  def add_or_reuse_constraint(self, label, ctype):
    """Adds a constraint to this LP, or reuses one if the label exists.

     If the constraint is already present, simply checks it's the same type as
     previously specified.

    Args:
      label: a label to assign to this constraint
      ctype: the constraint type (in CONS_TYPE_{LEQ,GEQ,EQ})
    """
    assert ctype in self._valid_constraint_types
    cons = self._cons.get(label)
    if cons is not None:
      # Do not re-add, but ensure it's the same type
      assert cons.ctype == ctype
      return
    if ctype == CONS_TYPE_LEQ or ctype == CONS_TYPE_GEQ:
      cons = _Constraint(len(self._leq_cons_list), ctype)
      self._cons[label] = cons
      self._leq_cons_list.append(cons)
    elif ctype == CONS_TYPE_EQ:
      cons = _Constraint(len(self._eq_cons_list), ctype)
      self._cons[label] = cons
      self._eq_cons_list.append(cons)
    else:
      assert False, "Unknown constraint type"

  def set_obj_coeff(self, var_label, coeff):
    """Sets a coefficient of a variable in the objective."""
    self._obj_coeffs[var_label] = coeff

  def set_cons_coeff(self, cons_label, var_label, coeff):
    """Sets a coefficient of a constraint in the LP."""
    self._cons[cons_label].coeffs[var_label] = coeff

  def add_to_cons_coeff(self, cons_label, var_label, add_coeff):
    """Sets a coefficient of a constraint in the LP."""
    val = self._cons[cons_label].coeffs.get(var_label)
    if val is None:
      val = 0
    self._cons[cons_label].coeffs[var_label] = val + add_coeff

  def set_cons_rhs(self, cons_label, value):
    """Sets the right-hand side of a constraint."""
    self._cons[cons_label].rhs = value

  def get_var_id(self, label):
    var = self._vars.get(label)
    assert var is not None
    return var.vid

  def get_num_cons(self):
    return len(self._leq_cons_list), len(self._eq_cons_list)

  def solve(self, solver=None):
    """Solves the LP.

    Args:
      solver: the solver to use ('blas', 'lapack', 'glpk'). Defaults to None,
        which then uses the cvxopt internal default.

    Returns:
      The solution as a dict of var label -> value, one for each variable.
    """
    # From http://cvxopt.org/userguide/coneprog.html#linear-programming,
    # CVXOPT uses the formulation:
    #    minimize: c^t x
    #       s.t.   Gx <= h
    #              Ax = b
    #
    # Here:
    #  - x is the vector the variables
    #  - c is the vector of objective coefficients
    #  - G is the matrix of LEQ (and GEQ) constraint coefficients
    #  - h is the vector or right-hand side values of the LEQ/GEQ constraints
    #  - A is the matrix of equality constraint coefficients
    #  - b is the vector of right-hand side values of the equality constraints
    #
    # This function builds these sparse matrices from the information it has
    # gathered, flipping signs where necessary, and adding equality constraints
    # for the upper and lower bounds of variables. It then calls the cvxopt
    # solver and maps back the values.
    num_vars = len(self._var_list)
    num_eq_cons = len(self._eq_cons_list)
    num_leq_cons = len(self._leq_cons_list)
    for var in self._var_list:
      if var.lb is not None:
        num_leq_cons += 1
      if var.ub is not None:
        num_leq_cons += 1
    # Make the matrices (some need to be dense).
    c = cvxopt.matrix([0.0] * num_vars)
    h = cvxopt.matrix([0.0] * num_leq_cons)
    g_mat = cvxopt.spmatrix([], [], [], (num_leq_cons, num_vars))
    a_mat = None
    b = None
    if num_eq_cons > 0:
      a_mat = cvxopt.spmatrix([], [], [], (num_eq_cons, num_vars))
      b = cvxopt.matrix([0.0] * num_eq_cons)
    # Objective coefficients: c
    for var_label in self._obj_coeffs:
      value = self._obj_coeffs[var_label]
      vid = self._vars[var_label].vid
      if self._objective == OBJ_MAX:
        c[vid] = -value  # negate the value because it's a max
      else:
        c[vid] = value  # min objective matches cvxopt
    # Inequality constraints: G, h
    row = 0
    for cons in self._leq_cons_list:
      # If it's >= then need to negate all coeffs and the rhs
      if cons.rhs is not None:
        h[row] = cons.rhs if cons.ctype == CONS_TYPE_LEQ else -cons.rhs
      for var_label in cons.coeffs:
        value = cons.coeffs[var_label]
        vid = self._vars[var_label].vid
        g_mat[(row, vid)] = value if cons.ctype == CONS_TYPE_LEQ else -value
      row += 1
    # Inequality constraints: variables upper and lower bounds
    for var in self._var_list:
      if var.lb is not None:  # x_i >= lb has to be -x_i <= -lb
        g_mat[(row, var.vid)] = -1.0
        h[row] = -var.lb
        row += 1
      if var.ub is not None:  # x_i <= ub
        g_mat[(row, var.vid)] = 1.0
        h[row] = var.ub
        row += 1
    # Equality constraints: A, b
    if num_eq_cons > 0:
      row = 0
      for cons in self._eq_cons_list:
        b[row] = cons.rhs if cons.rhs is not None else 0.0
        for var_label in cons.coeffs:
          value = cons.coeffs[var_label]
          vid = self._vars[var_label].vid
          a_mat[(row, vid)] = value
        row += 1
    # Solve!
    if num_eq_cons > 0:
      sol = cvxopt.solvers.lp(c, g_mat, h, a_mat, b, solver=solver)
    else:
      sol = cvxopt.solvers.lp(c, g_mat, h, solver=solver)
    return sol["x"]


def solve_zero_sum_matrix_game(game):
  """Solves a matrix game by using linear programming.

  Args:
    game: a pyspiel MatrixGame

  Returns:
    A 4-tuple containing:
      - p0_sol (array-like): probability distribution over row actions
      - p1_sol (array-like): probability distribution over column actions
      - p0_sol_value, expected value to the first player
      - p1_sol_value, expected value to the second player
  """

  # Solving a game for player i (e.g. row player) requires finding a mixed
  # policy over player i's pure strategies (actions) such that a value of the
  # mixed strategy against every opponent pure strategy is maximized.
  #
  # For more detail, please refer to Sec 4.1 of Shoham & Leyton-Brown, 2009:
  # Multiagent Systems: Algorithmic, Game-Theoretic, and Logical Foundations
  # http://www.masfoundations.org/mas.pdf
  #
  # For the row player the LP looks like:
  #    max V
  #     st. sigma_a1 \dot col_0 >= V
  #         sigma_a2 \dot col_1 >= V
  #              .
  #              .
  #         sigma_am \cot col_n >= V
  #         for all i, sigma_ai >= 0
  #         sigma \dot 1 = 1
  assert isinstance(game, pyspiel.MatrixGame)
  assert game.get_type().information == pyspiel.GameType.Information.ONE_SHOT
  assert game.get_type().utility == pyspiel.GameType.Utility.ZERO_SUM
  num_rows = game.num_rows()
  num_cols = game.num_cols()
  cvxopt.solvers.options["show_progress"] = False

  # First, do the row player (player 0).
  lp0 = LinearProgram(OBJ_MAX)
  for r in range(num_rows):  # one var per action / pure strategy
    lp0.add_or_reuse_variable(r, lb=0)
  lp0.add_or_reuse_variable(num_rows)  # V
  lp0.set_obj_coeff(num_rows, 1.0)  # max V
  for c in range(num_cols):
    lp0.add_or_reuse_constraint(c, CONS_TYPE_GEQ)
    for r in range(num_rows):
      lp0.set_cons_coeff(c, r, game.player_utility(0, r, c))
    lp0.set_cons_coeff(c, num_rows, -1.0)  # -V >= 0
  lp0.add_or_reuse_constraint(num_cols + 1, CONS_TYPE_EQ)
  lp0.set_cons_rhs(num_cols + 1, 1.0)
  for r in range(num_rows):
    lp0.set_cons_coeff(num_cols + 1, r, 1.0)
  sol = lp0.solve()
  p0_sol = sol[:-1]
  p0_sol_val = sol[-1]

  # Now, the column player (player 1).
  lp1 = LinearProgram(OBJ_MAX)
  for c in range(num_cols):  # one var per action / pure strategy
    lp1.add_or_reuse_variable(c, lb=0)
  lp1.add_or_reuse_variable(num_cols)  # V
  lp1.set_obj_coeff(num_cols, 1)  # max V
  for r in range(num_rows):
    lp1.add_or_reuse_constraint(r, CONS_TYPE_GEQ)
    for c in range(num_cols):
      lp1.set_cons_coeff(r, c, game.player_utility(1, r, c))
    lp1.set_cons_coeff(r, num_cols, -1.0)  # -V >= 0
  lp1.add_or_reuse_constraint(num_rows + 1, CONS_TYPE_EQ)
  lp1.set_cons_rhs(num_rows + 1, 1.0)
  for c in range(num_cols):
    lp1.set_cons_coeff(num_rows + 1, c, 1.0)
  sol = lp1.solve()
  p1_sol = sol[:-1]
  p1_sol_val = sol[-1]

  return p0_sol, p1_sol, p0_sol_val, p1_sol_val


def is_dominated(action,
                 game_or_payoffs,
                 player,
                 mode=DOMINANCE_STRICT,
                 tol=1e-7,
                 return_mixture=False):
  """Determines whether a pure strategy is dominated by any mixture strategies.

  Args:
    action: index of an action for `player`
    game_or_payoffs: either a pyspiel matrix- or normal-form game, or a payoff
      tensor for `player` with ndim == number of players
    player: index of the player (an integer)
    mode: dominance criterion: strict, weak, or very weak
    tol: tolerance
    return_mixture: whether to return the dominating strategy if one exists

  Returns:
    If `return_mixture`:
      a dominating mixture strategy if one exists, or `None`.
      the strategy is provided as a 1D numpy array of mixture weights.
    Otherwise: True if a dominating strategy exists, False otherwise.
  """
  # For more detail, please refer to Sec 4.5.2 of Shoham & Leyton-Brown, 2009:
  # Multiagent Systems: Algorithmic, Game-Theoretic, and Logical Foundations
  # http://www.masfoundations.org/mas.pdf
  assert mode in (DOMINANCE_STRICT, DOMINANCE_VERY_WEAK, DOMINANCE_WEAK)
  payoffs = utils.game_payoffs_array(game_or_payoffs)[player] if isinstance(
      game_or_payoffs, pyspiel.NormalFormGame) else np.asfarray(game_or_payoffs)

  # Reshape payoffs so rows correspond to `player` and cols to the joint action
  # of all other players
  payoffs = np.moveaxis(payoffs, player, 0)
  payoffs = payoffs.reshape((payoffs.shape[0], -1))
  num_rows, num_cols = payoffs.shape

  cvxopt.solvers.options["show_progress"] = False
  cvxopt.solvers.options["maxtol"] = tol
  cvxopt.solvers.options["feastol"] = tol
  lp = LinearProgram(OBJ_MAX)

  # One var for every row probability, fixed to 0 if inactive
  for r in range(num_rows):
    if r == action:
      lp.add_or_reuse_variable(r, lb=0, ub=0)
    else:
      lp.add_or_reuse_variable(r, lb=0)

  # For the strict LP we normalize the payoffs to be strictly positive
  if mode == DOMINANCE_STRICT:
    to_subtract = payoffs.min() - 1
  else:
    to_subtract = 0
    # For non-strict LPs the probabilities must sum to 1
    lp.add_or_reuse_constraint(num_cols, CONS_TYPE_EQ)
    lp.set_cons_rhs(num_cols, 1)
    for r in range(num_rows):
      if r != action:
        lp.set_cons_coeff(num_cols, r, 1)

  # The main dominance constraint
  for c in range(num_cols):
    lp.add_or_reuse_constraint(c, CONS_TYPE_GEQ)
    lp.set_cons_rhs(c, payoffs[action, c] - to_subtract)
    for r in range(num_rows):
      if r != action:
        lp.set_cons_coeff(c, r, payoffs[r, c] - to_subtract)

  if mode == DOMINANCE_STRICT:
    # Minimize sum of probabilities
    for r in range(num_rows):
      if r != action:
        lp.set_obj_coeff(r, -1)
    mixture = lp.solve()
    if mixture is not None and np.sum(mixture) < 1 - tol:
      mixture = np.squeeze(mixture, 1) / np.sum(mixture)
    else:
      mixture = None

  if mode == DOMINANCE_VERY_WEAK:
    # Check feasibility
    mixture = lp.solve()
    if mixture is not None:
      mixture = np.squeeze(mixture, 1)

  if mode == DOMINANCE_WEAK:
    # Check feasibility and whether there's any advantage
    for r in range(num_rows):
      lp.set_obj_coeff(r, payoffs[r].sum())
    mixture = lp.solve()
    if mixture is not None:
      mixture = np.squeeze(mixture, 1)
      if (np.dot(mixture, payoffs) - payoffs[action]).sum() <= tol:
        mixture = None

  return mixture if return_mixture else (mixture is not None)


def _pure_dominated_from_advantages(advantages, mode, tol=1e-7):
  if mode == DOMINANCE_STRICT:
    return (advantages > tol).all(1)
  if mode == DOMINANCE_WEAK:
    return (advantages >= -tol).all(1) & (advantages.sum(1) > tol)
  if mode == DOMINANCE_VERY_WEAK:
    return (advantages >= -tol).all(1)


def iterated_dominance(game_or_payoffs, mode, tol=1e-7):
  """Reduces a strategy space using iterated dominance.

  See: http://www.smallparty.com/yoram/classes/principles/nash.pdf

  Args:
    game_or_payoffs: either a pyspiel matrix- or normal-form game, or a payoff
      tensor of dimension `num_players` + 1. First dimension is the player,
      followed by the actions of all players, e.g. a 3x3 game (2 players) has
      dimension [2,3,3].
    mode: DOMINANCE_STRICT, DOMINANCE_WEAK, or DOMINANCE_VERY_WEAK
    tol: tolerance

  Returns:
    A tuple (`reduced_game`, `live_actions`).
    * if `game_or_payoffs` is an instance of `pyspiel.MatrixGame`, so is
      `reduced_game`; otherwise `reduced_game` is a payoff tensor.
    * `live_actions` is a tuple of length `num_players`, where
      `live_actions[player]` is a boolean vector of shape `num_actions`;
       `live_actions[player][action]` is `True` if `action` wasn't dominated for
       `player`.
  """
  payoffs = utils.game_payoffs_array(game_or_payoffs) if isinstance(
      game_or_payoffs, pyspiel.NormalFormGame) else np.asfarray(game_or_payoffs)
  live_actions = [
      np.ones(num_actions, np.bool) for num_actions in payoffs.shape[1:]
  ]
  progress = True
  while progress:
    progress = False
    # trying faster method first
    for method in ("pure", "mixed"):
      if progress:
        continue
      for player, live in enumerate(live_actions):
        if live.sum() == 1:
          # one action is dominant
          continue

        # discarding all dominated opponent actions
        payoffs_live = payoffs[player]
        for opponent in range(payoffs.shape[0]):
          if opponent != player:
            payoffs_live = payoffs_live.compress(live_actions[opponent],
                                                 opponent)

        # reshaping to (player_actions, joint_opponent_actions)
        payoffs_live = np.moveaxis(payoffs_live, player, 0)
        payoffs_live = payoffs_live.reshape((payoffs_live.shape[0], -1))

        for action in range(live.size):
          if not live[action]:
            continue
          if method == "pure":
            # mark all actions that `action` dominates
            advantage = payoffs_live[action] - payoffs_live
            dominated = _pure_dominated_from_advantages(advantage, mode, tol)
            dominated[action] = False
            dominated &= live
            if dominated.any():
              progress = True
              live &= ~dominated
              if live.sum() == 1:
                break
          if method == "mixed":
            # test if `action` is dominated by a mixed policy
            mixture = is_dominated(
                live[:action].sum(),
                payoffs_live[live],
                0,
                mode,
                tol,
                return_mixture=True)
            if mixture is None:
              continue
            # if it is, mark any other actions dominated by that policy
            progress = True
            advantage = mixture.dot(payoffs_live[live]) - payoffs_live[live]
            dominated = _pure_dominated_from_advantages(advantage, mode, tol)
            dominated[mixture > tol] = False
            assert dominated[live[:action].sum()]
            live.put(live.nonzero()[0], ~dominated)
            if live.sum() == 1:
              break

  for player, live in enumerate(live_actions):
    payoffs = payoffs.compress(live, player + 1)

  if isinstance(game_or_payoffs, pyspiel.MatrixGame):
    return pyspiel.MatrixGame(game_or_payoffs.get_type(),
                              game_or_payoffs.get_parameters(), [
                                  game_or_payoffs.row_action_name(action)
                                  for action in live_actions[0].nonzero()[0]
                              ], [
                                  game_or_payoffs.col_action_name(action)
                                  for action in live_actions[1].nonzero()[0]
                              ], *payoffs), live_actions
  else:
    return payoffs, live_actions


# TODO(author5): add a function for sequential games using sequence-form LPs.



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python implementation of the CFR-BR algorithm."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import cfr
from open_spiel.python.algorithms import exploitability
import pyspiel

# pylint: disable=protected-access
_CFRSolverBase = cfr._CFRSolverBase
_update_current_policy = cfr._update_current_policy
_apply_regret_matching_plus_reset = cfr._apply_regret_matching_plus_reset
# pylint: enable=protected-access


class CFRBRSolver(_CFRSolverBase):
  """Implements the Counterfactual Regret Minimization (CFR-BR) algorithm.

  This is Counterfactual Regret Minimization against Best Response, from
  Michael Johanson and al., 2012, Finding Optimal Abstract Strategies in
  Extensive-Form Games,
  https://poker.cs.ualberta.ca/publications/AAAI12-cfrbr.pdf).

  The algorithm
  computes an approximate Nash policy for n-player zero-sum games, but the
  implementation is currently restricted to 2-player.

  It uses an exact Best Response and full tree traversal.

  One iteration for a n-player game consists of the following:

  - Compute the BR of each player against the rest of the players.
  - Then, for each player p sequentially (from player 0 to N-1):
    - Compute the conterfactual reach probabilities and action values for player
      p, playing against the set of the BR for all other players.
    - Update the player `p` policy using these values.

  CFR-BR should converge with high probability (see the paper), but we can also
  compute the time-averaged strategy.

  The implementation reuses the `action_values_vs_best_response` module and
  thus uses TabularPolicies. This will run only for smallish games.
  """

  def __init__(self, game, linear_averaging=False, regret_matching_plus=False):
    # pyformat: disable
    """Initializer.

    Args:
      game: The `pyspiel.Game` to run on.
      linear_averaging: Whether to use linear averaging, i.e.
        cumulative_policy[info_state][action] += (
          iteration_number * reach_prob * action_prob)

        or not:

        cumulative_policy[info_state][action] += reach_prob * action_prob
      regret_matching_plus: Whether to use Regret Matching+:
        cumulative_regrets = max(cumulative_regrets + regrets, 0)
        or simply regret matching:
        cumulative_regrets = cumulative_regrets + regrets
    """
    # pyformat: enable
    if game.num_players() != 2:
      raise ValueError("Game {} does not have {} players.".format(game, 2))

    assert game.get_type().dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL, (
        "CFR requires sequential games. If you're trying to run it " +
        "on a simultaneous (or normal-form) game, please first transform it " +
        "using turn_based_simultaneous_game.")

    super(CFRBRSolver, self).__init__(
        game,
        alternating_updates=True,
        linear_averaging=linear_averaging,
        regret_matching_plus=regret_matching_plus)

    self._best_responses = {i: None for i in range(game.num_players())}

  def _compute_best_responses(self):
    """Computes each player best-response against the pool of other players."""

    def policy_fn(state):
      key = state.information_state_string()
      return self._get_infostate_policy(key)

    current_policy = policy.tabular_policy_from_callable(self._game, policy_fn)

    for player_id in range(self._game.num_players()):
      self._best_responses[player_id] = exploitability.best_response(
          self._game, current_policy, player_id)

  def evaluate_and_update_policy(self):
    """Performs a single step of policy evaluation and policy improvement."""
    self._iteration += 1

    self._compute_best_responses()

    for player in range(self._num_players):
      # We do not use policies, to not have to call `state.information_state`
      # several times (in here and within policy).
      policies = []
      for p in range(self._num_players):
        # pylint: disable=g-long-lambda
        policies.append(
            lambda infostate_str, p=p:
            {self._best_responses[p]["best_response_action"][infostate_str]: 1})
        # pylint: enable=g-long-lambda
      policies[player] = self._get_infostate_policy

      self._compute_counterfactual_regret_for_player(
          state=self._root_node,
          policies=policies,
          reach_probabilities=np.ones(self._num_players + 1),
          player=player)

      if self._regret_matching_plus:
        _apply_regret_matching_plus_reset(self._info_state_nodes)
    _update_current_policy(self._current_policy, self._info_state_nodes)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example algorithm to get all states from a game."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


def _get_subgames_states(state, all_states, depth_limit, depth,
                         include_terminals, include_chance_states, to_string,
                         stop_if_encountered):
  """Extract non-chance states for a subgame into the all_states dict."""
  if state.is_terminal():
    if include_terminals:
      # Include if not already present and then terminate recursion.
      state_str = to_string(state)
      if state_str not in all_states:
        all_states[state_str] = state.clone()
    return

  if depth > depth_limit >= 0:
    return

  if not state.is_chance_node() or include_chance_states:
    # Add only if not already present
    state_str = to_string(state)
    if state_str not in all_states:
      all_states[state_str] = state.clone()
    else:
      # We already saw this one. Stop the recursion if the flag is set
      if stop_if_encountered:
        return

  for action in state.legal_actions():
    state_for_search = state.child(action)
    _get_subgames_states(state_for_search, all_states, depth_limit, depth + 1,
                         include_terminals, include_chance_states, to_string,
                         stop_if_encountered)


def get_all_states(game,
                   depth_limit=-1,
                   include_terminals=True,
                   include_chance_states=False,
                   to_string=lambda s: s.history_str(),
                   stop_if_encountered=True):
  """Gets all states in the game, indexed by their string representation.

  For small games only! Useful for methods that solve the  games explicitly,
  i.e. value iteration. Use this default implementation with caution as it does
  a recursive tree walk of the game and could easily fill up memory for larger
  games or games with long horizons.

  Currently only works for sequential games.

  Arguments:
    game: The game to analyze, as returned by `load_game`.
    depth_limit: How deeply to analyze the game tree. Negative means no limit, 0
      means root-only, etc.
    include_terminals: If True, include terminal states.
    include_chance_states: If True, include chance node states.
    to_string: The serialization function. We expect this to be
      `lambda s: s.history_str()` as this enforces perfect recall, but for
        historical reasons, using `str` is also supported, but the goal is to
        remove this argument.
    stop_if_encountered: if this is set, do not keep recursively adding states
      if this state is already in the list. This allows support for games that
      have cycles.

  Returns:
    A `dict` with `to_string(state)` keys and `pyspiel.State` values containing
    all states encountered traversing the game tree up to the specified depth.
  """
  # Get the root state.
  state = game.new_initial_state()
  all_states = dict()

  # Then, do a recursive tree walk to fill up the map.
  _get_subgames_states(
      state=state,
      all_states=all_states,
      depth_limit=depth_limit,
      depth=0,
      include_terminals=include_terminals,
      include_chance_states=include_chance_states,
      to_string=to_string,
      stop_if_encountered=stop_if_encountered)

  if not all_states:
    raise ValueError("GetSubgameStates returned 0 states!")

  return all_states



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ephemeral Value Adjustment example: https://arxiv.org/abs/1810.08163."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
from absl import logging

import tensorflow.compat.v1 as tf

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import eva
from open_spiel.python.algorithms import exploitability
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_episodes", 1000, "Number of iterations")
flags.DEFINE_string("game_name", "kuhn_poker", "Name of the game")


class JointPolicy(policy.Policy):
  """Joint policy to be evaluated."""

  def __init__(self, agents):
    self._agents = agents

  def action_probabilities(self, state, player_id=None):
    cur_player = state.current_player()
    return self._agents[cur_player].action_probabilities(state)


def main(unused_argv):
  logging.info("Loading %s", FLAGS.game_name)
  env = rl_environment.Environment(FLAGS.game_name)
  num_players = env.num_players
  num_actions = env.action_spec()["num_actions"]
  state_size = env.observation_spec()["info_state"][0]
  eva_agents = []
  with tf.Session() as sess:
    for player in range(num_players):
      eva_agents.append(
          eva.EVAAgent(
              sess,
              env,
              player,
              state_size,
              num_actions,
              embedding_network_layers=(64, 32),
              embedding_size=12,
              learning_rate=1e-4,
              mixing_parameter=0.5,
              memory_capacity=int(1e6),
              discount_factor=1.0,
              epsilon_start=1.0,
              epsilon_end=0.1,
              epsilon_decay_duration=int(1e6)))
    sess.run(tf.global_variables_initializer())
    time_step = env.reset()
    for _ in range(FLAGS.num_episodes):
      while not time_step.last():
        current_player = time_step.observations["current_player"]
        current_agent = eva_agents[current_player]
        step_out = current_agent.step(time_step)
        time_step = env.step([step_out.action])

      for agent in eva_agents:
        agent.step(time_step)

    game = pyspiel.load_game(FLAGS.game_name)
    joint_policy = JointPolicy(eva_agents)
    conv = exploitability.nash_conv(game, joint_policy)
    logging.info("EVA in '%s' - NashConv: %s", FLAGS.game_name, conv)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Functions to manipulate game playthoughs.

Used by examples/playthrough.py and tests/playthrough_test.py.
"""

import os
import re
import numpy as np

from open_spiel.python.games import tic_tac_toe
from open_spiel.python.observation import make_observation
import pyspiel


def _load_game(game_string):
  """Loads a game, including special-cases for Python-implemented games."""
  if game_string == "python_tic_tac_toe":
    return tic_tac_toe.TicTacToeGame()
  else:
    return pyspiel.load_game(game_string)


def _escape(x):
  """Returns a newline-free backslash-escaped version of the given string."""
  x = x.replace("\\", R"\\")
  x = x.replace("\n", R"\n")
  return x


def _format_value(v):
  """Format a single value."""
  if v == 0:
    return "◯"
  elif v == 1:
    return "◉"
  else:
    return ValueError("Values must all be 0 or 1")


def _format_vec(vec):
  return "".join(_format_value(v) for v in vec)


def _format_matrix(mat):
  return np.char.array([_format_vec(row) for row in mat])


def _format_tensor(tensor, tensor_name, max_cols=120):
  """Formats a tensor in an easy-to-view format as a list of lines."""
  if ((tensor.shape == (0,)) or (len(tensor.shape) > 3) or
      not np.logical_or(tensor == 0, tensor == 1).all()):
    vec = ", ".join(str(round(v, 5)) for v in tensor.ravel())
    return ["{} = [{}]".format(tensor_name, vec)]
  elif len(tensor.shape) == 1:
    return ["{}: {}".format(tensor_name, _format_vec(tensor))]
  elif len(tensor.shape) == 2:
    if len(tensor_name) + tensor.shape[0] + 2 < max_cols:
      lines = ["{}: {}".format(tensor_name, _format_vec(tensor[0]))]
      prefix = " " * (len(tensor_name) + 2)
    else:
      lines = ["{}:".format(tensor_name), _format_vec(tensor[0])]
      prefix = ""
    for row in tensor[1:]:
      lines.append(prefix + _format_vec(row))
    return lines
  elif len(tensor.shape) == 3:
    lines = ["{}:".format(tensor_name)]
    rows = []
    for m in tensor:
      formatted_matrix = _format_matrix(m)
      if (not rows) or (len(rows[-1][0] + formatted_matrix[0]) + 2 > max_cols):
        rows.append(formatted_matrix)
      else:
        rows[-1] = rows[-1] + "  " + formatted_matrix
    for i, big_row in enumerate(rows):
      if i > 0:
        lines.append("")
      for row in big_row:
        lines.append("".join(row))
    return lines


def playthrough(game_string, action_sequence, alsologtostdout=False):
  """Returns a playthrough of the specified game as a single text.

  Actions are selected uniformly at random, including chance actions.

  Args:
    game_string: string, e.g. 'markov_soccer', with possible optional params,
      e.g. 'go(komi=4.5,board_size=19)'.
    action_sequence: A (possibly partial) list of action choices to make.
    alsologtostdout: Whether to also print the trace to stdout. This can be
      useful when an error occurs, to still be able to get context information.
  """
  lines = playthrough_lines(game_string, alsologtostdout, action_sequence)
  return "\n".join(lines) + "\n"


def format_shapes(d):
  """Returns a string representing the shapes of a dict of tensors."""
  if len(d) == 1:
    return str(list(d[min(d)].shape))
  else:
    return ", ".join(f"{key}: {list(value.shape)}" for key, value in d.items())


def playthrough_lines(game_string, alsologtostdout=False, action_sequence=None):
  """Returns a playthrough of the specified game as a list of lines.

  Actions are selected uniformly at random, including chance actions.

  Args:
    game_string: string, e.g. 'markov_soccer' or 'kuhn_poker(players=4)'.
    alsologtostdout: Whether to also print the trace to stdout. This can be
      useful when an error occurs, to still be able to get context information.
    action_sequence: A (possibly partial) list of action choices to make.
  """
  lines = []
  action_sequence = action_sequence or []
  if alsologtostdout:

    def add_line(v):
      print(v)
      lines.append(v)
  else:
    add_line = lines.append

  game = _load_game(game_string)
  add_line("game: {}".format(game_string))
  seed = np.random.randint(2**32 - 1)
  game_type = game.get_type()

  default_observation = None
  try:
    default_observation = make_observation(game)
  except (RuntimeError, ValueError):
    pass

  infostate_observation = None
  # TODO(author11) reinstate this restriction
  # if game_type.information in (pyspiel.IMPERFECT_INFORMATION,
  #                              pyspiel.ONE_SHOT):
  try:
    infostate_observation = make_observation(
        game, pyspiel.IIGObservationType(perfect_recall=True))
  except (RuntimeError, ValueError):
    pass

  public_observation = None
  try:
    public_observation = make_observation(
        game,
        pyspiel.IIGObservationType(
            public_info=True,
            perfect_recall=False,
            private_info=pyspiel.PrivateInfoType.NONE))
  except (RuntimeError, ValueError):
    pass

  private_observation = None
  try:
    private_observation = make_observation(
        game,
        pyspiel.IIGObservationType(
            public_info=False,
            perfect_recall=False,
            private_info=pyspiel.PrivateInfoType.SINGLE_PLAYER))
  except (RuntimeError, ValueError):
    pass

  add_line("")
  add_line("GameType.chance_mode = {}".format(game_type.chance_mode))
  add_line("GameType.dynamics = {}".format(game_type.dynamics))
  add_line("GameType.information = {}".format(game_type.information))
  add_line("GameType.long_name = {}".format('"{}"'.format(game_type.long_name)))
  add_line("GameType.max_num_players = {}".format(game_type.max_num_players))
  add_line("GameType.min_num_players = {}".format(game_type.min_num_players))
  add_line("GameType.parameter_specification = {}".format("[{}]".format(
      ", ".join('"{}"'.format(param)
                for param in sorted(game_type.parameter_specification)))))
  add_line("GameType.provides_information_state_string = {}".format(
      game_type.provides_information_state_string))
  add_line("GameType.provides_information_state_tensor = {}".format(
      game_type.provides_information_state_tensor))
  add_line("GameType.provides_observation_string = {}".format(
      game_type.provides_observation_string))
  add_line("GameType.provides_observation_tensor = {}".format(
      game_type.provides_observation_tensor))
  add_line("GameType.provides_factored_observation_string = {}".format(
      game_type.provides_factored_observation_string))
  add_line("GameType.reward_model = {}".format(game_type.reward_model))
  add_line("GameType.short_name = {}".format('"{}"'.format(
      game_type.short_name)))
  add_line("GameType.utility = {}".format(game_type.utility))

  add_line("")
  add_line("NumDistinctActions() = {}".format(game.num_distinct_actions()))
  add_line("PolicyTensorShape() = {}".format(game.policy_tensor_shape()))
  add_line("MaxChanceOutcomes() = {}".format(game.max_chance_outcomes()))
  add_line("GetParameters() = {{{}}}".format(",".join(
      "{}={}".format(key, _escape(str(value)))
      for key, value in sorted(game.get_parameters().items()))))
  add_line("NumPlayers() = {}".format(game.num_players()))
  add_line("MinUtility() = {:.5}".format(game.min_utility()))
  add_line("MaxUtility() = {:.5}".format(game.max_utility()))
  try:
    utility_sum = game.utility_sum()
  except RuntimeError:
    utility_sum = None
  add_line("UtilitySum() = {}".format(utility_sum))
  if infostate_observation and infostate_observation.tensor is not None:
    add_line("InformationStateTensorShape() = {}".format(
        format_shapes(infostate_observation.dict)))
    add_line("InformationStateTensorLayout() = {}".format(
        game.information_state_tensor_layout()))
    add_line("InformationStateTensorSize() = {}".format(
        len(infostate_observation.tensor)))
  if default_observation and default_observation.tensor is not None:
    add_line("ObservationTensorShape() = {}".format(
        format_shapes(default_observation.dict)))
    add_line("ObservationTensorLayout() = {}".format(
        game.observation_tensor_layout()))
    add_line("ObservationTensorSize() = {}".format(
        len(default_observation.tensor)))
  add_line("MaxGameLength() = {}".format(game.max_game_length()))
  add_line('ToString() = "{}"'.format(str(game)))

  players = list(range(game.num_players()))
  state = game.new_initial_state()
  state_idx = 0
  rng = np.random.RandomState(seed)

  while True:
    add_line("")
    add_line("# State {}".format(state_idx))
    for line in str(state).splitlines():
      add_line("# {}".format(line).rstrip())
    add_line("IsTerminal() = {}".format(state.is_terminal()))
    add_line("History() = {}".format([int(a) for a in state.history()]))
    add_line('HistoryString() = "{}"'.format(state.history_str()))
    add_line("IsChanceNode() = {}".format(state.is_chance_node()))
    add_line("IsSimultaneousNode() = {}".format(state.is_simultaneous_node()))
    add_line("CurrentPlayer() = {}".format(state.current_player()))
    if infostate_observation:
      for player in players:
        s = infostate_observation.string_from(state, player)
        if s is not None:
          add_line(f'InformationStateString({player}) = "{_escape(s)}"')
    if infostate_observation and infostate_observation.tensor is not None:
      for player in players:
        infostate_observation.set_from(state, player)
        for name, tensor in infostate_observation.dict.items():
          label = f"InformationStateTensor({player})"
          label += f".{name}" if name != "info_state" else ""
          lines += _format_tensor(tensor, label)
    if default_observation:
      for player in players:
        s = default_observation.string_from(state, player)
        if s is not None:
          add_line(f'ObservationString({player}) = "{_escape(s)}"')
    if public_observation:
      s = public_observation.string_from(state, 0)
      if s is not None:
        add_line('PublicObservationString() = "{}"'.format(_escape(s)))
      for player in players:
        s = private_observation.string_from(state, player)
        if s is not None:
          add_line(f'PrivateObservationString({player}) = "{_escape(s)}"')
    if default_observation and default_observation.tensor is not None:
      for player in players:
        default_observation.set_from(state, player)
        for name, tensor in default_observation.dict.items():
          label = f"ObservationTensor({player})"
          label += f".{name}" if name != "observation" else ""
          lines += _format_tensor(tensor, label)
    if game_type.chance_mode == pyspiel.GameType.ChanceMode.SAMPLED_STOCHASTIC:
      add_line('SerializeState() = "{}"'.format(_escape(state.serialize())))
    if not state.is_chance_node():
      add_line("Rewards() = {}".format(state.rewards()))
      add_line("Returns() = {}".format(state.returns()))
    if state.is_terminal():
      break
    if state.is_chance_node():
      # In Python 2 and Python 3, the default number of decimal places displayed
      # is different. Thus, we hardcode a ".12" which is Python 2 behaviour.
      add_line("ChanceOutcomes() = [{}]".format(", ".join(
          "{{{}, {:.12f}}}".format(outcome, prob)
          for outcome, prob in state.chance_outcomes())))
    if state.is_simultaneous_node():
      for player in players:
        add_line("LegalActions({}) = [{}]".format(
            player, ", ".join(str(x) for x in state.legal_actions(player))))
      for player in players:
        add_line("StringLegalActions({}) = [{}]".format(
            player, ", ".join('"{}"'.format(state.action_to_string(player, x))
                              for x in state.legal_actions(player))))
      if state_idx < len(action_sequence):
        actions = action_sequence[state_idx]
      else:
        actions = [rng.choice(state.legal_actions(pl)) for pl in players]
      add_line("")
      add_line("# Apply joint action [{}]".format(
          format(", ".join(
              '"{}"'.format(state.action_to_string(player, action))
              for player, action in enumerate(actions)))))
      add_line("actions: [{}]".format(", ".join(
          str(action) for action in actions)))
      state.apply_actions(actions)
    else:
      add_line("LegalActions() = [{}]".format(", ".join(
          str(x) for x in state.legal_actions())))
      add_line("StringLegalActions() = [{}]".format(", ".join(
          '"{}"'.format(state.action_to_string(state.current_player(), x))
          for x in state.legal_actions())))
      if state_idx < len(action_sequence):
        action = action_sequence[state_idx]
      else:
        action = rng.choice(state.legal_actions())
      add_line("")
      add_line('# Apply action "{}"'.format(
          state.action_to_string(state.current_player(), action)))
      add_line("action: {}".format(action))
      state.apply_action(action)
    state_idx += 1
  return lines


def content_lines(lines):
  """Return lines with content."""
  return [line for line in lines if line and line[0] == "#"]


def _playthrough_params(lines):
  """Returns the playthrough parameters from a playthrough record.

  Args:
    lines: The playthrough as a list of lines.

  Returns:
    A `dict` with entries:
      game_string: string, e.g. 'markov_soccer'.
      action_sequence: a list of action choices made in the playthrough.
    Suitable for passing to playthrough to re-generate the playthrough.

  Raises:
    ValueError if the playthrough is not valid.
  """
  params = {"action_sequence": []}
  for line in lines:
    match_game = re.match(r"^game: (.*)$", line)
    match_action = re.match(r"^action: (.*)$", line)
    match_actions = re.match(r"^actions: \[(.*)\]$", line)
    if match_game:
      params["game_string"] = match_game.group(1)
    if match_action:
      params["action_sequence"].append(int(match_action.group(1)))
    if match_actions:
      params["action_sequence"].append(
          [int(x) for x in match_actions.group(1).split(", ")])
  if "game_string" in params:
    return params
  raise ValueError("Could not find params")


def _read_playthrough(filename):
  """Returns the content and the parsed arguments of a playthrough file."""
  with open(filename, "r", encoding="utf-8") as f:
    original = f.read()
  kwargs = _playthrough_params(original.splitlines())
  return original, kwargs


def replay(filename):
  """Re-runs the playthrough in the specified file. Returns (original, new)."""
  original, kwargs = _read_playthrough(filename)
  return (original, playthrough(**kwargs))


def update_path(path, shard_index=0, num_shards=1):
  """Regenerates all playthroughs in the path."""
  for filename in sorted(os.listdir(path))[shard_index::num_shards]:
    try:
      original, kwargs = _read_playthrough(os.path.join(path, filename))
      try:
        _load_game(kwargs["game_string"])
      except pyspiel.SpielError as e:
        if "Unknown game" in str(e):
          print("[Skipped] Skipping game ", filename, " as ",
                kwargs["game_string"], " is not available.")
          continue
      new = playthrough(**kwargs)
      if original == new:
        print("          {}".format(filename))
      else:
        with open(os.path.join(path, filename), "w") as f:
          f.write(new)
        print("Updated {}".format(filename))
    except Exception as e:  # pylint: disable=broad-except
      print("{} failed: {}".format(filename, e))
      raise



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements ResponseGraphUCB algorithm from the below paper.

  "Multiagent Evaluation under Incomplete Information" (Rowland et al., 2019)
  See https://arxiv.org/abs/1909.09849 for details.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import functools
import itertools
import operator
import random

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import scipy.stats


class ResponseGraphUCB(object):
  """ResponseGraphUCB sampler class."""

  def __init__(self,
               game,
               exploration_strategy='uniform-exhaustive',
               confidence_method='ucb-standard',
               delta=0.01,
               ucb_eps=0,
               per_payoff_confidence=True,
               time_dependent_delta=False):
    """Initializes ResponseGraphUCB instance.

    Assumes that all payoffs fall in the interval [0,1].

    Args:
      game: an instance of the BernoulliGameSampler class.
      exploration_strategy: string specifying the exploration strategy.
      confidence_method: string specifying the confidence method.
      delta: float specifying the UCB delta parameter.
      ucb_eps: float specifying the UCB epsilon parameter.
      per_payoff_confidence: bool specifying whether confidence level applies
        on a per-payoff basis, or to all payoffs simultaneously.
      time_dependent_delta: bool specifying whether the confidence parameter
        varies with the number of interactions so that a union bound holds.
    """
    self.exploration_strategy = exploration_strategy
    self.confidence_method = confidence_method
    self.ucb_eps = ucb_eps
    self.G = game  # pylint: disable=invalid-name
    self.per_payoff_confidence = per_payoff_confidence
    self.time_dependent_delta = time_dependent_delta
    if self.per_payoff_confidence:
      self._delta = delta
    else:
      self._delta = delta / (
          self.G.n_players *
          functools.reduce(operator.mul, self.G.strategy_spaces, 1))

    # Compute the graph
    self.V = list(  # pylint: disable=invalid-name
        itertools.product(*[range(smax) for smax in self.G.strategy_spaces]))
    self.E = []  # pylint: disable=invalid-name
    for v in self.V:
      adj_strats = [
          list(range(v[k] + 1, self.G.strategy_spaces[k]))
          for k in range(self.G.n_players)
      ]
      for k in range(self.G.n_players):
        for new_s in adj_strats[k]:
          second_vertex = list(v)
          second_vertex[k] = new_s
          second_vertex = tuple(second_vertex)
          self.E.append((v, second_vertex))
    self.count_history = {v: [] for v in self.V}
    self.total_interactions = 0

  def delta(self, k, s):
    """Returns the confidence parameter for a given player and profile."""
    if not self.time_dependent_delta:
      return self._delta
    else:
      return self._delta * (6 / (np.pi**2 * self.count[k][s] **2))

  def initialise_mean_and_count(self):
    """Initializes means and counts for all response graph profiles."""
    self.mu = [
        np.zeros(tuple(self.G.strategy_spaces)) for _ in range(self.G.n_players)
    ]
    self.count = [
        np.zeros(tuple(self.G.strategy_spaces)) for _ in range(self.G.n_players)
    ]

  def update_mean_and_count(self, strat_profile, game_outcome):
    """Updates means and counts for strat_profile given game_outcome."""
    self.total_interactions += 1
    for k in range(self.G.n_players):
      self.mu[k][strat_profile] *= self.count[k][strat_profile]
      self.mu[k][strat_profile] += game_outcome[k]
      self.count[k][strat_profile] += 1
      self.mu[k][strat_profile] /= self.count[k][strat_profile]

    for s in self.V:
      self.count_history[s].append(self.count[0][s] /
                                   float(self.total_interactions))

  def _find_focal_coord(self, s1, s2):
    num_deviations = tuple(s1[l] != s2[l] for l in range(len(s1)))
    assert np.sum(num_deviations) == 1, ('Invalid profile pair s1, s2: ({},{}).'
                                         'Exactly one player should'
                                         'deviate!'.format(s1, s2))
    return np.argmax(num_deviations)

  def _initialise_queue_uniform(self):
    self.remaining_edges = copy.deepcopy(self.E)

  def _add_to_queue_uniform(self, edges_removed):
    """Adds edge to sampling queue using uniform sampling."""
    for e in edges_removed:
      self.remaining_edges.remove(e)
    self.profile_queue.append(
        random.choice(random.choice(self.remaining_edges)))

  def _initialise_queue_uniform_exhaustive(self):
    self.edge_order = copy.deepcopy(self.E)
    random.shuffle(self.edge_order)

  def _add_to_queue_uniform_exhaustive(self, edges_removed):
    """Adds edge to sampling queue using uniform-exhausitive sampling."""
    for e in edges_removed:
      self.edge_order.remove(e)
    self.profile_queue.append(random.choice(self.edge_order[0]))

  def _initialise_queue_valence_weighted(self):
    self.vertex_valences = {
        v: np.sum(self.G.strategy_spaces) - self.G.n_players for v in self.V
    }
    self.sum_valences = sum(self.vertex_valences.values())

  def _add_to_queue_valence_weighted(self, edges_removed):
    """Adds edge to sampling queue using valence-weighted sampling."""
    # Deal with removed edges
    for e in edges_removed:
      for s in e:
        self.vertex_valences[s] -= 1
        self.sum_valences -= 1

    # Calculate probabilities
    probs = np.array([self.vertex_valences[v]**2 for v in self.V])
    probs = probs / np.sum(probs)
    s_ix = np.random.choice(np.arange(len(self.V)), p=probs)
    self.profile_queue.append(self.V[s_ix])

  def _initialise_queue_count_weighted(self):
    # Keep track of which vertices have non-zero valence in graph
    self.vertex_valences = {
        v: np.sum(self.G.strategy_spaces) - self.G.n_players for v in self.V
    }
    self.sum_valences = sum(self.vertex_valences.values())

  def _add_to_queue_count_weighted(self, edges_removed):
    """Adds edge to sampling queue using count-weighted sampling."""
    # Update vertex valences
    for e in edges_removed:
      for s in e:
        self.vertex_valences[s] -= 1
        self.sum_valences -= 1
    # Check counts
    eligible_vertices = {
        v: self.count[0][v] for v in self.V if self.vertex_valences[v] != 0
    }
    strat = min(eligible_vertices, key=eligible_vertices.get)
    self.profile_queue.append(strat)

  def initialise_queue(self):
    """Initializes sampling queue."""
    self.edges_remaining = copy.deepcopy(self.E)
    if self.exploration_strategy == 'uniform':
      self._initialise_queue_uniform()
    elif self.exploration_strategy == 'uniform-exhaustive':
      self._initialise_queue_uniform_exhaustive()
    elif self.exploration_strategy == 'valence-weighted':
      self._initialise_queue_valence_weighted()
    elif self.exploration_strategy == 'count-weighted':
      self._initialise_queue_count_weighted()
    else:
      raise ValueError('Did not recognise exploration strategy: {}'.format(
          self.exploration_strategy))

    self.profile_queue = []

  def add_to_queue(self, removed):
    """Update the sampling queue and the list of resolved edges.

    Args:
      removed: the list of edges resolved in the previous round, which should be
        removed from the sampling list in subsequent rounds.
    """
    if self.exploration_strategy == 'uniform':
      self._add_to_queue_uniform(removed)
    elif self.exploration_strategy == 'uniform-exhaustive':
      self._add_to_queue_uniform_exhaustive(removed)
    elif self.exploration_strategy == 'valence-weighted':
      self._add_to_queue_valence_weighted(removed)
    elif self.exploration_strategy == 'count-weighted':
      self._add_to_queue_count_weighted(removed)
    else:
      raise ValueError('Did not recognise exploration strategy: {}'.format(
          self.exploration_strategy))

  def evaluate_strategy_profile(self, yield_outcomes=False):
    """Evaluates a strategy profile on the sampling queue.

    Specifically, this:
      1. Removes a strategy profile from the queue.
      2. Evaluates it.
      3. Updates internal statistics.
      4. Adjusts list of strategy profiles whose statistics have been updated
         since last confidence bound check.

    Args:
      yield_outcomes: set True to yield the outcomes as well.

    Yields:
      s: profile evaluated.
      game_outcome: outcomes (player payoffs) for profile s.
    """
    if self.profile_queue:
      s = self.profile_queue.pop(0)
      if s not in self.active_strategy_profiles:
        self.active_strategy_profiles.append(s)
      game_outcome = self.G.observe_result(s)
      if yield_outcomes:
        yield s, game_outcome
      self.update_mean_and_count(s, game_outcome)

  def _ucb_standard_factor(self, s, k):
    return np.sqrt(np.log(2 / self.delta(k, s)) / (2 * self.count[k][s]))

  def _bernoulli_upper(self, p, n, delta):
    """Returns upper confidence bound for proportion p successes of n trials.

    Uses exact Clopper-Pearson interval.

    Args:
      p: proportion of successes.
      n: number of trials.
      delta: confidence parameter.
    """
    if p > 1 - 1e-6:
      return 1.
    else:
      upper = scipy.stats.beta.ppf(1. - delta / 2, p * n + 1, n - p * n)
      return upper

  def _bernoulli_lower(self, p, n, delta):
    """Returns lower confidence bound for proportion p successes of n trials.

    Uses exact Clopper-Pearson interval.

    Args:
      p: proportion of successes.
      n: number of trials.
      delta: confidence parameter.
    """
    if p < 1e-6:
      return 0.
    else:
      lower = scipy.stats.beta.ppf(delta / 2, p * n, n - p * n + 1)
      return lower

  def _ucb(self, s, k):
    """Returns k-th player's payoff upper-confidence-bound given profile s."""
    if self.confidence_method == 'ucb-standard':
      ucb_factor = self._ucb_standard_factor(s, k)
      return self.mu[k][s] + ucb_factor
    elif self.confidence_method == 'ucb-standard-relaxed':
      ucb_factor = self._ucb_standard_factor(s, k) - self.ucb_eps
      return self.mu[k][s] + ucb_factor
    elif self.confidence_method == 'clopper-pearson-ucb':
      return self._bernoulli_upper(self.mu[k][s], self.count[k][s],
                                   self.delta(k, s))
    elif self.confidence_method == 'clopper-pearson-ucb-relaxed':
      return self._bernoulli_upper(self.mu[k][s], self.count[k][s],
                                   self.delta(k, s)) - self.ucb_eps
    else:
      raise ValueError('Did not recognise confidence method {}'.format(
          self.confidence_method))

  def _lcb(self, s, k):
    """Returns k-th player's payoff lower-confidence-bound given profile s."""
    if self.confidence_method == 'ucb-standard':
      ucb_factor = self._ucb_standard_factor(s, k)
      return self.mu[k][s] - ucb_factor
    elif self.confidence_method == 'ucb-standard-relaxed':
      ucb_factor = self._ucb_standard_factor(s, k) + self.ucb_eps
      return self.mu[k][s] - ucb_factor
    elif self.confidence_method == 'clopper-pearson-ucb':
      return self._bernoulli_lower(self.mu[k][s], self.count[k][s],
                                   self.delta(k, s))
    elif self.confidence_method == 'clopper-pearson-ucb-relaxed':
      return self._bernoulli_lower(self.mu[k][s], self.count[k][s],
                                   self.delta(k, s)) + self.ucb_eps
    else:
      raise ValueError('Did not recognise confidence method {}'.format(
          self.confidence_method))

  def ucb_check(self, e):
    """Conducts a UCB check on response graph edge e.

    Specifically, given edge e connecting two strategy profiles s1 and s2, this:
      1. Determines the dominating strategy.
      2. Checks whether the payoff_UCB(worse_strategy) is less than
        the payoff_LCB of the better strategy; if this is true, the confidence
        intervals are disjoint, and the edge e is considered 'resolved'.

    Args:
      e: response graph edge.

    Returns:
      A bool indicating whether the edge is resolved,
      and also a tuple specifying the worse and better strategies.
    """

    s1, s2 = e
    k = self._find_focal_coord(s1, s2)
    if self.mu[k][s1] > self.mu[k][s2]:
      better_strat = s1
      worse_strat = s2
    else:
      better_strat = s2
      worse_strat = s1

    ucb = self._ucb(worse_strat, k)
    lcb = self._lcb(better_strat, k)

    return (ucb < lcb), (worse_strat, better_strat)

  def check_confidence(self):
    """Returns the edges that are 'resolved' given a confidence bound check."""
    edges_to_check = []

    for e in self.edges_remaining:
      for s in self.active_strategy_profiles:
        if s in e:
          if e not in edges_to_check:
            edges_to_check.append(e)

    edges_removed = []
    for e in edges_to_check:
      removed, ordered_edge = self.ucb_check(e)
      if removed:
        edges_removed.append(e)
        self.edges_remaining.remove(e)
        self.directed_edges.append(ordered_edge)

    self.active_strategy_profiles = []

    return edges_removed

  def real_edge_direction(self, e):
    s1, s2 = e
    k = self._find_focal_coord(s1, s2)
    if self.G.means[k][s1] > self.G.means[k][s2]:
      return (s2, s1)
    else:
      return (s1, s2)

  def construct_real_graph(self):
    directed_edges = []
    for e in self.E:
      ordered_edge = self.real_edge_direction(e)
      directed_edges.append(ordered_edge)

    return self._construct_digraph(directed_edges)

  def compute_graph(self):
    for e in self.E:
      s1, s2 = e[0], e[1]
      k = self._find_focal_coord(s1, s2)
      if self.mu[k][s1] > self.mu[k][s2]:
        directed_edge = (s2, s1)
      else:
        directed_edge = (s1, s2)
      if directed_edge not in self.directed_edges:
        self.directed_edges.append(directed_edge)

  def forced_exploration(self):
    for v in self.V:
      game_outcome = self.G.observe_result(v)
      self.update_mean_and_count(v, game_outcome)

  def run(self, verbose=True, max_total_iterations=50000):
    """Runs the ResponseGraphUCB algorithm."""
    self.verbose = verbose

    # Upper bounds on number of evaluations
    self.max_total_iterations = max_total_iterations

    self.initialise_mean_and_count()
    self.directed_edges = []
    self.active_strategy_profiles = []
    self.initialise_queue()

    # Forced initial exploration
    self.forced_exploration()

    # Keep evaluating nodes until check method declares that we're finished
    iterations = 0
    edges_resolved_this_round = []
    while self.total_interactions < max_total_iterations:
      # Add nodes to queue
      self.add_to_queue(removed=edges_resolved_this_round)

      # Evaluate the nodes and log results
      for v, _ in self.evaluate_strategy_profile():
        if verbose:
          print(v)

      # Recompute confidence bounds, eliminate, stop etc.
      edges_resolved_this_round = self.check_confidence()

      if not self.edges_remaining:
        break
      iterations += 1

    # Fill in missing edges if max iters reached without resolving all edges
    self.compute_graph()

    # Compute objects to be returned
    if verbose:
      total_steps = self.compute_total_steps()
      print('\nTotal steps taken = {}'.format(total_steps))
    results = {}
    results['interactions'] = int(np.sum(self.count[0]))
    graph = self._construct_digraph(self.directed_edges)
    results['graph'] = graph
    return results

  def compute_total_steps(self):
    return int(np.sum(self.count[0]))

  def _construct_digraph(self, edges):
    graph = nx.DiGraph()
    graph.add_nodes_from(self.V)
    for e in edges:
      graph.add_edge(e[0], e[1])
    return graph

  def _plot_errorbars_2x2x2(self, x, y, xerr, yerr, fmt):
    """Plots ResponseGraph with error bars for a 2-player 2x2 game."""

    # plt.errorbar does not accept list of colors, so plot twice
    for i_strat in [0, 1]:
      if xerr[i_strat] is None:
        plt.errorbar(
            x=x[i_strat],
            y=y[i_strat],
            yerr=np.reshape(yerr[:, i_strat], (2, 1)),
            markerfacecolor='b',
            ecolor='b',
            fmt=fmt,
            zorder=0)
      elif yerr[i_strat] is None:
        plt.errorbar(
            x=x[i_strat],
            y=y[i_strat],
            xerr=np.reshape(xerr[:, i_strat], (2, 1)),
            markerfacecolor='b',
            ecolor='b',
            fmt=fmt,
            zorder=0)
      else:
        raise ValueError()

  def visualise_2x2x2(self, real_values, graph):
    """Plots summary of ResponseGraphUCB for a 2-player 2x2 game."""
    _, axes = plt.subplots(3, 3, figsize=(10, 10),
                           gridspec_kw={'width_ratios': [1, 2, 1],
                                        'height_ratios': [1, 2, 1]})
    axes[0, 0].axis('off')
    axes[0, 2].axis('off')
    axes[2, 0].axis('off')
    axes[2, 2].axis('off')

    # (0,0) vs. (0,1)
    plt.sca(axes[0, 1])
    s1 = (0, 0)
    s2 = (0, 1)
    self._plot_errorbars_2x2x2(
        x=[0, 1],
        y=[self.mu[1][s1], self.mu[1][s2]],
        xerr=[None, None],
        yerr=np.array([[self.mu[1][s1] - self._lcb(s1, 1),
                        self.mu[1][s2] - self._lcb(s2, 1)],
                       [self._ucb(s1, 1) - self.mu[1][s1],
                        self._ucb(s2, 1) - self.mu[1][s2]]]),
        fmt='o')
    plt.scatter([0, 1], [real_values[1, 0, 0], real_values[1, 0, 1]],
                color='red',
                zorder=1)
    plt.tick_params(axis='both', which='major', labelsize=14)
    plt.tick_params(axis='both', which='minor', labelsize=14)
    plt.xticks([])
    plt.yticks([0, 0.5, 1])
    plt.gca().set_yticklabels(['0', '', '1'])
    plt.gca().yaxis.set_ticks_position('left')
    plt.gca().grid(True)
    plt.ylim(0, 1)

    # (0,0) vs. (1,0)
    plt.sca(axes[1, 0])
    s1 = (1, 0)
    s2 = (0, 0)
    self._plot_errorbars_2x2x2(
        x=[self.mu[0][s1], self.mu[0][s2]],
        y=[0, 1],
        xerr=np.array([[self.mu[0][s1] - self._lcb(s1, 0),
                        self.mu[0][s2] - self._lcb(s2, 0)],
                       [self._ucb(s1, 0) - self.mu[0][s1],
                        self._ucb(s2, 0) - self.mu[0][s2]]]),
        yerr=[None, None],
        fmt='o')
    plt.scatter([real_values[0, 1, 0], real_values[0, 0, 0]], [0, 1],
                color='red',
                zorder=1)
    plt.tick_params(axis='both', which='major', labelsize=14)
    plt.tick_params(axis='both', which='minor', labelsize=14)
    plt.xticks([0, 0.5, 1])
    plt.gca().set_xticklabels(['0', '', '1'])
    plt.gca().xaxis.set_ticks_position('bottom')
    plt.gca().grid(True)
    plt.yticks([])
    plt.xlim(0, 1)

    # (0,1) vs. (1,1)
    plt.sca(axes[1, 2])
    s1 = (1, 1)
    s2 = (0, 1)
    self._plot_errorbars_2x2x2(
        x=[self.mu[0][s1], self.mu[0][s2]],
        y=[0, 1],
        xerr=np.array([[self.mu[0][s1] - self._lcb(s1, 0),
                        self.mu[0][s2] - self._lcb(s2, 0)],
                       [self._ucb(s1, 0) - self.mu[0][s1],
                        self._ucb(s2, 0) - self.mu[0][s2]]]),
        yerr=[None, None],
        fmt='o')
    plt.scatter([real_values[0, 1, 1], real_values[0, 0, 1]], [0, 1],
                color='red',
                zorder=1)
    plt.tick_params(axis='both', which='major', labelsize=14)
    plt.tick_params(axis='both', which='minor', labelsize=14)
    plt.xticks([0, 0.5, 1])
    plt.gca().set_xticklabels(['0', '', '1'])
    plt.gca().xaxis.set_ticks_position('top')
    plt.yticks([])
    plt.gca().grid(True)
    plt.xlim(0, 1)

    # (1,0) vs. (1,1)
    plt.sca(axes[2, 1])
    s1 = (1, 0)
    s2 = (1, 1)
    self._plot_errorbars_2x2x2(
        x=[0, 1],
        y=[self.mu[1][s1], self.mu[1][s2]],
        xerr=[None, None],
        yerr=np.array([[self.mu[1][s1] - self._lcb(s1, 1),
                        self.mu[1][s2] - self._lcb(s2, 1)],
                       [self._ucb(s1, 1) - self.mu[1][s1],
                        self._ucb(s2, 1) - self.mu[1][s2]]]),
        fmt='o')
    plt.scatter([0, 1], [real_values[1, 1, 0], real_values[1, 1, 1]],
                color='red',
                zorder=1)
    plt.tick_params(axis='both', which='major', labelsize=14)
    plt.tick_params(axis='both', which='minor', labelsize=14)
    plt.xticks([])
    plt.yticks([0, 0.5, 1])
    plt.gca().set_yticklabels(['0', '', '1'])
    plt.gca().yaxis.set_ticks_position('right')
    plt.gca().grid(True)
    plt.ylim(0, 1)
    self.plot_graph(graph, subplot=True, axes=axes)  # Chart in the middle

  def plot_graph(self, graph, subplot=False, axes=None):
    """Plots the response graph."""
    if subplot:
      plt.sca(axes[1, 1])
      axes[1, 1].axis('off')
    else:
      plt.figure(figsize=(5, 5))
    if len(graph.nodes) == 4:
      pos = {(0, 0): [0, 1], (0, 1): [1, 1], (1, 0): [0, 0], (1, 1): [1, 0]}
    else:
      pos = nx.circular_layout(graph)
    nx.draw_networkx_nodes(
        graph, pos, node_size=1800, node_color='w', edgecolors='k')
    nx.draw_networkx_edges(
        graph,
        pos,
        node_size=1800,
        edge_color='k',
        arrowstyle='->',
        arrowsize=10,
        width=3)
    nx.draw_networkx_labels(self.G, pos, {x: x for x in self.V}, font_size=14)

  def visualise_count_history(self, figsize=(5, 2)):
    """Plots the sampling count history for each strategy profile."""
    plt.figure(figsize=figsize)
    data = []
    labels = []
    for v in self.V:
      print(v)
      labels.append(v)
      data.append(self.count_history[v])
    pal = plt.get_cmap('Dark2').colors
    plt.stackplot(
        np.arange(1, self.total_interactions + 1),
        np.array(data),
        labels=labels,
        colors=pal)
    plt.ylim(top=1, bottom=0)
    plt.xlabel('Interactions')
    plt.ylabel('Proportions')

    # Shrink current axis
    ax = plt.gca()
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.67, box.height])
    plt.xlim(1, self.total_interactions)
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), ncol=1)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is replaces google's gfile used for network storage.

A more complete public version of gfile:
https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/platform/gfile.py
"""

import os

# pylint: disable=invalid-name
Exists = os.path.exists
IsDirectory = os.path.isdir
ListDir = os.listdir
MakeDirs = os.makedirs
Open = open



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various visualization tools for Alpha-Rank.

All equations and variable names correspond to the following paper:
  https://arxiv.org/abs/1903.01373

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import logging

try:
  import matplotlib.patches as patches  # pylint: disable=g-import-not-at-top
  import matplotlib.patheffects as PathEffects  # pylint: disable=g-import-not-at-top
  import matplotlib.pyplot as plt  # pylint: disable=g-import-not-at-top
except ImportError as e:
  logging.info("If your tests failed with the error 'ImportError: No module "
               "named functools_lru_cache', this is a known bug in matplotlib "
               "and there is a workaround (run sudo apt install "
               "python-backports.functools-lru-cache. See: "
               "https://github.com/matplotlib/matplotlib/issues/9344.")
  raise ImportError(str(e))

import networkx as nx  # pylint: disable=g-import-not-at-top
import numpy as np

from open_spiel.python.egt import utils


class NetworkPlot(object):
  """A class for visualizing the Alpha-Rank interaction network."""

  def __init__(self,
               payoff_tables,
               rhos,
               rho_m,
               pi,
               state_labels,
               num_top_profiles=None):
    """Initializes a network plotting object.

    Args:
      payoff_tables: List of game payoff tables, one for each agent identity.
        Each payoff_table may be either a 2D numpy array, or a
        _PayoffTableInterface object.
      rhos: Fixation probabilities.
      rho_m: Neutral fixation probability.
      pi: Stationary distribution of fixation Markov chain defined by rhos.
      state_labels: Labels corresponding to Markov states. For the
        single-population case, state_labels should be a list of pure strategy
        names. For the multi-population case, it
                    should be a dict with (key,value) pairs: (population
                      index,list of strategy names)
      num_top_profiles: Set to (int) to show only the graph nodes corresponding
        to the top k elements of stationary distribution, or None to show all.
    """
    self.fig = plt.figure(figsize=(10, 10))
    self.num_populations = len(payoff_tables)
    payoffs_are_hpt_format = utils.check_payoffs_are_hpt(payoff_tables)
    self.num_strats_per_population =\
      utils.get_num_strats_per_population(payoff_tables, payoffs_are_hpt_format)
    self.rhos = rhos
    self.rho_m = rho_m
    self.pi = pi
    self.num_profiles = len(pi)
    self.state_labels = state_labels
    self.first_run = True
    self.num_top_profiles = num_top_profiles

    if self.num_top_profiles:
      # More than total number of strats requested for plotting
      if self.num_top_profiles > self.num_profiles:
        self.num_top_profiles = self.num_profiles
      # Skip the bottom num_profiles-k stationary strategies.
      self.nodes_to_skip = list(self.pi.argsort()[:self.num_profiles-\
                                                  self.num_top_profiles])
    else:
      self.nodes_to_skip = []

    self._reset_cycle_counter()

  def _reset_cycle_counter(self):
    self.i_cycle_to_show = -1

  def _draw_network(self):
    """Draws the NetworkX object representing the underlying graph."""
    plt.clf()

    if self.num_populations == 1:
      node_sizes = 5000
      node_border_width = 1.
    else:
      node_sizes = 15000
      node_border_width = 3.

    vmin, vmax = 0, np.max(self.pi) + 0.1

    nx.draw_networkx_nodes(
        self.g,
        self.pos,
        node_size=node_sizes,
        node_color=self.node_colors,
        edgecolors="k",
        cmap=plt.cm.Blues,
        vmin=vmin,
        vmax=vmax,
        linewidths=node_border_width)

    nx.draw_networkx_edges(
        self.g,
        self.pos,
        node_size=node_sizes,
        arrowstyle="->",
        arrowsize=10,
        edge_color=self.edge_colors,
        edge_cmap=plt.cm.Blues,
        width=5)

    nx.draw_networkx_edge_labels(self.g, self.pos, edge_labels=self.edge_labels)

    if self.num_populations > 1:
      subnode_separation = 0.1
      subgraph = nx.Graph()
      for i_population in range(self.num_populations):
        subgraph.add_node(i_population)

    for i_strat_profile in self.g:
      x, y = self.pos[i_strat_profile]
      if self.num_populations == 1:
        node_text = "$\\pi_{" + self.state_labels[i_strat_profile] + "}=$"
        node_text += str(np.round(self.pi[i_strat_profile], decimals=2))
      else:
        node_text = ""  # No text for multi-population case as plot gets messy
      txt = plt.text(
          x,
          y,
          node_text,
          horizontalalignment="center",
          verticalalignment="center",
          fontsize=12)
      txt.set_path_effects(
          [PathEffects.withStroke(linewidth=3, foreground="w")])

      if self.num_populations > 1:
        sub_pos = nx.circular_layout(subgraph)
        subnode_labels = dict()
        strat_profile = utils.get_strat_profile_from_id(
            self.num_strats_per_population, i_strat_profile)
        for i_population in subgraph.nodes():
          i_strat = strat_profile[i_population]
          subnode_labels[i_population] = "$s^{" + str(i_population + 1) + "}="
          subnode_labels[i_population] +=\
              self.state_labels[i_population][i_strat] + "$"
          # Adjust the node positions generated by NetworkX's circular_layout(),
          # such that the node for the 1st strategy starts on the left.
          sub_pos[i_population] = (-sub_pos[i_population] * subnode_separation +
                                   self.pos[i_strat_profile])
        nx.draw(
            subgraph,
            pos=sub_pos,
            with_labels=True,
            width=0.,
            node_color="w",
            labels=subnode_labels,
            node_size=2500)

  def compute_and_draw_network(self):
    """Computes the various node/edge connections of the graph and draws it."""

    if np.max(self.rhos) < self.rho_m:
      print("All node-to-node fixation probabilities (not including self-cycles"
            " are lower than neutral. Thus, no graph will be drawn.")
      return

    self.g = nx.MultiDiGraph()
    self.edge_labels = {}
    self.edge_alphas = []
    rho_max = np.max(self.rhos / self.rho_m)
    rho_m_alpha = 0.1  # Transparency of neutral selection edges

    for i in range(self.num_profiles):
      for j in range(self.num_profiles):
        # Do not draw edge if any node involved is skipped
        if j not in self.nodes_to_skip and i not in self.nodes_to_skip:
          rate = self.rhos[i][j] / self.rho_m
          # Draws edges when fixation from one strategy to another occurs (i.e.,
          # rate > 1), or with fixation equal to neutral selection probability
          # (i.e., rate == 1). This is consistent with visualizations used in
          # finite-population literature.
          if rate > 1:
            # Compute alphas. Clip needed due to numerical precision.
            alpha = np.clip(rho_m_alpha + (1 - rho_m_alpha) * rate / rho_max,
                            None, 1.)
            self.g.add_edge(i, j, weight=alpha, label="{:.01f}".format(rate))
            self.edge_alphas.append(alpha)
          elif np.isclose(rate, 1):
            alpha = rho_m_alpha
            self.g.add_edge(i, j, weight=alpha, label="{:.01f}".format(rate))
            self.edge_alphas.append(alpha)
          # Label edges for non-self-loops with sufficient flowrate
          if i != j and rate > 1:
            edge_string = "$" + str(np.round(rate, decimals=2)) + "\\rho_m$"
          else:
            edge_string = ""
          self.edge_labels[(i, j)] = edge_string

    # MultiDiGraph nodes are not ordered, so order the node colors accordingly
    self.node_colors = [self.pi[node] for node in self.g.nodes()]

    self.cycles = list(nx.simple_cycles(self.g))
    self.num_cycles = len(self.cycles)

    # Color the edges of cycles if user requested it
    if self.i_cycle_to_show >= 0:
      all_cycle_edges = [
          zip(nodes, (nodes[1:] + nodes[:1])) for nodes in self.cycles
      ]
      cur_cycle_edges = all_cycle_edges[self.i_cycle_to_show]
      self.edge_colors = []
      for u, v in self.g.edges():
        if (u, v) in cur_cycle_edges:
          self.edge_colors.append([1., 0., 0.])
        else:
          self.edge_colors.append([1. - self.g[u][v][0]["weight"]] * 3)
    else:
      self.edge_colors = [
          [1. - self.g[u][v][0]["weight"]] * 3 for u, v in self.g.edges()
      ]
      self.edge_alphas = [self.g[u][v][0]["weight"] for u, v in self.g.edges()]

    ax = plt.gca()

    # Centered circular pose
    self.pos = nx.layout.circular_layout(self.g)
    all_x = [node_pos[0] for node, node_pos in self.pos.items()]
    all_y = [node_pos[1] for node, node_pos in self.pos.items()]
    min_x = np.min(all_x)
    max_x = np.max(all_x)
    min_y = np.min(all_y)
    max_y = np.max(all_y)
    for node, node_pos in self.pos.items():
      node_pos[0] -= (max_x + min_x) / 2
      node_pos[1] -= (max_y + min_y) / 2

    # Rendering
    self._draw_network()
    if self.first_run:
      ax.autoscale_view()
    ax.set_axis_off()
    ax.set_aspect("equal")
    plt.ylim(-1.3, 1.3)
    plt.xlim(-1.3, 1.3)
    if self.first_run:
      self.first_run = False
      plt.axis("off")
      plt.show()


def _draw_pie(ax,
              ratios,
              colors,
              x_center=0,
              y_center=0,
              size=100,
              clip_on=True,
              zorder=0):
  """Plots a pie chart.

  Args:
    ax: plot axis.
    ratios: list indicating size of each pie slice, with elements summing to 1.
    colors: list indicating color of each pie slice.
    x_center: x coordinate of pie center.
    y_center: y coordinate of pie center.
    size: pie size.
    clip_on: control clipping of pie (e.g., to show it when it's out of axis).
    zorder: plot z order (e.g., to show pie on top of other plot elements).
  """
  xy = []
  start = 0.
  for ratio in ratios:
    x = [0] + np.cos(
        np.linspace(2 * np.pi * start, 2 * np.pi *
                    (start + ratio), 30)).tolist()
    y = [0] + np.sin(
        np.linspace(2 * np.pi * start, 2 * np.pi *
                    (start + ratio), 30)).tolist()
    xy.append(list(zip(x, y)))
    start += ratio

  for i, xyi in enumerate(xy):
    ax.scatter([x_center], [y_center],
               marker=(xyi, 0),
               s=size,
               facecolor=colors[i],
               edgecolors="none",
               clip_on=clip_on,
               zorder=zorder)


def generate_sorted_masses_strats(pi_list, curr_alpha_idx, strats_to_go):
  """Generates a sorted list of (mass, strats) tuples.

  Args:
    pi_list: List of stationary distributions, pi
    curr_alpha_idx: Index in alpha_list for which to start clustering
    strats_to_go: List of strategies that still need to be ordered

  Returns:
    Sorted list of (mass, strats) tuples.
  """
  if curr_alpha_idx > 0:
    sorted_masses_strats = list()
    masses_to_strats = utils.cluster_strats(pi_list[curr_alpha_idx,
                                                    strats_to_go])

    for mass, strats in sorted(masses_to_strats.items(), reverse=True):
      if len(strats) > 1:
        to_append = generate_sorted_masses_strats(pi_list, curr_alpha_idx - 1,
                                                  strats)

        to_append = [(mass, [strats_to_go[s]
                             for s in strats_list])
                     for (mass, strats_list) in to_append]

        sorted_masses_strats.extend(to_append)
      else:
        sorted_masses_strats.append((mass, [
            strats_to_go[strats[0]],
        ]))

    return sorted_masses_strats
  else:
    to_return = sorted(
        utils.cluster_strats(pi_list[curr_alpha_idx, strats_to_go]).items(),
        reverse=True)
    to_return = [(mass, [strats_to_go[s]
                         for s in strats_list])
                 for (mass, strats_list) in to_return]
    return to_return


def plot_pi_vs_alpha(pi_list,
                     alpha_list,
                     num_populations,
                     num_strats_per_population,
                     strat_labels,
                     num_strats_to_label,
                     plot_semilogx=True,
                     xlabel=r"Ranking-intensity $\alpha$",
                     ylabel=r"Strategy mass in stationary distribution $\pi$",
                     legend_sort_clusters=False):
  """Plots stationary distributions, pi, against selection intensities, alpha.

  Args:
    pi_list: List of stationary distributions, pi.
    alpha_list: List of selection intensities, alpha.
    num_populations: The number of populations.
    num_strats_per_population: List of the number of strategies per population.
    strat_labels: Human-readable strategy labels.
    num_strats_to_label: The number of top strategies to label in the legend.
    plot_semilogx: Boolean set to enable/disable semilogx plot.
    xlabel: Plot xlabel.
    ylabel: Plot ylabel.
    legend_sort_clusters: If true, strategies in the same cluster are sorted in
      the legend according to orderings for earlier alpha values. Primarily for
      visualization purposes! Rankings for lower alpha values should be
      interpreted carefully.
  """

  # Cluster strategies for which the stationary distribution has similar masses
  masses_to_strats = utils.cluster_strats(pi_list[-1, :])

  # Set colors
  num_strat_profiles = np.shape(pi_list)[1]
  num_strats_to_label = min(num_strats_to_label, num_strat_profiles)
  cmap = plt.get_cmap("Paired")
  colors = [cmap(i) for i in np.linspace(0, 1, num_strat_profiles)]

  # Plots stationary distribution vs. alpha series
  plt.figure(facecolor="w")
  axes = plt.gca()

  legend_line_objects = []
  legend_labels = []

  rank = 1
  num_strats_printed = 0
  add_legend_entries = True

  if legend_sort_clusters:
    sorted_masses_strats = generate_sorted_masses_strats(
        pi_list, pi_list.shape[0] - 1, range(pi_list.shape[1]))
  else:
    sorted_masses_strats = sorted(masses_to_strats.items(), reverse=True)

  for mass, strats in sorted_masses_strats:
    for profile_id in strats:
      if num_populations == 1:
        strat_profile = profile_id
      else:
        strat_profile = utils.get_strat_profile_from_id(
            num_strats_per_population, profile_id)

      if plot_semilogx:
        series = plt.semilogx(
            alpha_list,
            pi_list[:, profile_id],
            color=colors[profile_id],
            linewidth=2)
      else:
        series = plt.plot(
            alpha_list,
            pi_list[:, profile_id],
            color=colors[profile_id],
            linewidth=2)

      if add_legend_entries:
        if num_strats_printed >= num_strats_to_label:
          # Placeholder blank series for remaining entries
          series = plt.semilogx(np.NaN, np.NaN, "-", color="none")
          label = "..."
          add_legend_entries = False
        else:
          label = utils.get_label_from_strat_profile(num_populations,
                                                     strat_profile,
                                                     strat_labels)
        legend_labels.append(label)
        legend_line_objects.append(series[0])
      num_strats_printed += 1
    rank += 1

  # Plots pie charts on far right of figure to indicate clusters of strategies
  # with identical rank
  for mass, strats in iter(masses_to_strats.items()):
    _draw_pie(
        axes,
        ratios=[1 / len(strats)] * len(strats),
        colors=[colors[i] for i in strats],
        x_center=alpha_list[-1],
        y_center=mass,
        size=200,
        clip_on=False,
        zorder=10)

  # Axes ymax set slightly above highest stationary distribution mass
  max_mass = np.amax(pi_list)
  axes_y_max = np.ceil(
      10. * max_mass) / 10  # Round upward to nearest first decimal
  axes_y_max = np.clip(axes_y_max, 0., 1.)

  # Plots a rectangle highlighting the rankings on the far right of the figure
  box_x_min = alpha_list[-1] * 0.7
  box_y_min = np.min(pi_list[-1, :]) - 0.05 * axes_y_max
  width = 0.7 * alpha_list[-1]
  height = np.max(pi_list[-1, :]) - np.min(
      pi_list[-1, :]) + 0.05 * axes_y_max * 2
  axes.add_patch(
      patches.Rectangle((box_x_min, box_y_min),
                        width,
                        height,
                        edgecolor="b",
                        facecolor=(1, 0, 0, 0),
                        clip_on=False,
                        linewidth=5,
                        zorder=20))

  # Plot formatting
  axes.set_xlim(np.min(alpha_list), np.max(alpha_list))
  axes.set_ylim([0.0, axes_y_max])
  axes.set_xlabel(xlabel)
  axes.set_ylabel(ylabel)
  axes.set_axisbelow(True)  # Axes appear below data series in terms of zorder

  # Legend on the right side of the current axis
  box = axes.get_position()
  axes.set_position([box.x0, box.y0, box.width * 0.8, box.height])
  axes.legend(
      legend_line_objects,
      legend_labels,
      loc="center left",
      bbox_to_anchor=(1.05, 0.5))
  plt.grid()
  plt.show()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compute exploitability of a uniform policy."""

from absl import app
from absl import flags

from open_spiel.python import policy
from open_spiel.python.algorithms import exploitability
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_string("game", "kuhn_poker", "Name of the game")


def main(_):
  game = pyspiel.load_game(FLAGS.game)
  expl = exploitability.exploitability(game, policy.UniformRandomPolicy(game))
  print("Exploitability: {}".format(expl))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Class of Optimization Oracles generating best response against opponents.

Oracles are as defined in (Lanctot et Al., 2017,
https://arxiv.org/pdf/1711.00832.pdf ), functions generating a best response
against a probabilistic mixture of opponents. This class implements the abstract
class of oracles, and a simple oracle using Evolutionary Strategy as
optimization method.
"""

import numpy as np


def strategy_sampler_fun(total_policies, probabilities_of_playing_policies):
  """Samples strategies according to distribution over them.

  Args:
    total_policies: List of lists of policies for each player.
    probabilities_of_playing_policies: List of numpy arrays representing the
      probability of playing a strategy.

  Returns:
    One sampled joint strategy.
  """
  policies_selected = []
  for k in range(len(total_policies)):
    selected_opponent = np.random.choice(
        total_policies[k],
        1,
        p=probabilities_of_playing_policies[k]).reshape(-1)[0]
    policies_selected.append(selected_opponent)
  return policies_selected


class AbstractOracle(object):
  """The abstract class representing oracles, a hidden optimization process."""

  def __init__(self,
               number_policies_sampled=100,
               **oracle_specific_kwargs):
    """Initialization method for oracle.

    Args:
      number_policies_sampled: Number of different opponent policies sampled
        during evaluation of policy.
      **oracle_specific_kwargs: Oracle specific args, compatibility
        purpose. Since oracles can vary so much in their implementation, no
        specific argument constraint is put on this function.
    """
    self._number_policies_sampled = number_policies_sampled
    self._kwargs = oracle_specific_kwargs

  def set_iteration_numbers(self, number_policies_sampled):
    """Changes the number of iterations used for computing episode returns.

    Args:
      number_policies_sampled: Number of different opponent policies sampled
        during evaluation of policy.
    """
    self._number_policies_sampled = number_policies_sampled

  def __call__(self, game, policy, total_policies, current_player,
               probabilities_of_playing_policies,
               **oracle_specific_execution_kwargs):
    """Call method for oracle, returns best response against a set of policies.

    Args:
      game: The game on which the optimization process takes place.
      policy: The current policy, in policy.Policy, from which we wish to start
        optimizing.
      total_policies: A list of all policy.Policy strategies used for training,
        including the one for the current player.
      current_player: Integer representing the current player.
      probabilities_of_playing_policies: A list of arrays representing, per
        player, the probabilities of playing each policy in total_policies for
        the same player.
      **oracle_specific_execution_kwargs: Other set of arguments, for
        compatibility purposes. Can for example represent whether to Rectify
        Training or not.
    """
    raise NotImplementedError("Calling Abstract class method.")

  def sample_episode(self, game, policies_selected):
    raise NotImplementedError("Calling Abstract class method.")

  def evaluate_policy(self, game, pol, total_policies, current_player,
                      probabilities_of_playing_policies,
                      strategy_sampler=strategy_sampler_fun,
                      **oracle_specific_execution_kwargs):
    """Evaluates a specific policy against a nash mixture of policies.

    Args:
      game: The game on which the optimization process takes place.
      pol: The current policy, in policy.Policy, from which we wish to start
        optimizing.
      total_policies: A list of all policy.Policy strategies used for training,
        including the one for the current player.
      current_player: Integer representing the current player.
      probabilities_of_playing_policies: A list of arrays representing, per
        player, the probabilities of playing each policy in total_policies for
        the same player.
      strategy_sampler: callable sampling strategy.
      **oracle_specific_execution_kwargs: Other set of arguments, for
        compatibility purposes. Can for example represent whether to Rectify
        Training or not.

    Returns:
      Average return for policy when played against policies_played_against.
    """
    del oracle_specific_execution_kwargs  # Unused.

    totals = 0
    count = 0
    for _ in range(self._number_policies_sampled):
      policies_selected = strategy_sampler(total_policies,
                                           probabilities_of_playing_policies)
      policies_selected[current_player] = pol

      new_return = self.sample_episode(
          game,
          policies_selected)[current_player]

      totals += new_return
      count += 1

    # Avoid the 0 / 0 case.
    return totals / max(1, count)




# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.evaluate_bots."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import evaluate_bots
from open_spiel.python.bots import uniform_random
from open_spiel.python.bots.policy import PolicyBot
import pyspiel


GAME = pyspiel.load_game("kuhn_poker")


def policy_bots():
  random_policy = policy.UniformRandomPolicy(GAME)

  py_bot = PolicyBot(0, np.random.RandomState(4321), random_policy)
  cpp_bot = pyspiel.make_policy_bot(
      GAME, 1, 1234,
      policy.python_policy_to_pyspiel_policy(random_policy.to_tabular()))

  return [py_bot, cpp_bot]


class EvaluateBotsTest(parameterized.TestCase):

  @parameterized.parameters([([
      pyspiel.make_uniform_random_bot(0, 1234),
      uniform_random.UniformRandomBot(1, np.random.RandomState(4321))
  ],), (policy_bots(),)])
  def test_cpp_vs_python(self, bots):
    results = np.array([
        evaluate_bots.evaluate_bots(GAME.new_initial_state(), bots, np.random)
        for _ in range(10000)
    ])
    average_results = np.mean(results, axis=0)
    np.testing.assert_allclose(average_results, [0.125, -0.125], atol=0.1)

  def test_random_vs_stateful(self):
    game = pyspiel.load_game("tic_tac_toe")
    bots = [
        pyspiel.make_stateful_random_bot(game, 0, 1234),
        uniform_random.UniformRandomBot(1, np.random.RandomState(4321))
    ]
    for _ in range(1000):
      evaluate_bots.evaluate_bots(game.new_initial_state(), bots, np.random)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Play a game, selecting random moves, and save what we see.

This can be used to check by hand the behaviour of a game, and also
as the basis for test cases.

Example usage:
```
playthrough --game kuhn_poker --params players=3
```
"""

from absl import app
from absl import flags
from absl import logging

from open_spiel.python.algorithms import generate_playthrough

FLAGS = flags.FLAGS

flags.DEFINE_string(
    "game", "kuhn_poker", "Name of the game, with optional parameters, e.g. "
    "'kuhn_poker' or 'go(komi=4.5,board_size=19)'.")
flags.DEFINE_string("output_file", None, "Where to write the data to.")
flags.DEFINE_list("actions", None,
                  "A (possibly partial) list of action choices to make.")

flags.DEFINE_string("update_path", None,
                    "If set, regenerates all playthroughs in the path.")
flags.DEFINE_bool(
    "alsologtostdout", False,
    "If True, the trace will be written to std-out while it "
    "is being constructed (in addition to the usual behavior).")
flags.DEFINE_integer("shard", 0, "The shard to update.")
flags.DEFINE_integer("num_shards", 1, "How many shards to use for updates.")


def main(unused_argv):
  if FLAGS.update_path:
    generate_playthrough.update_path(FLAGS.update_path, FLAGS.shard,
                                     FLAGS.num_shards)
  else:
    if not FLAGS.game:
      raise ValueError("Must specify game")
    actions = FLAGS.actions
    if actions is not None:
      actions = [int(x) for x in actions]
    text = generate_playthrough.playthrough(
        FLAGS.game, actions, alsologtostdout=FLAGS.alsologtostdout)
    if FLAGS.output_file:
      with open(FLAGS.output_file, "w") as f:
        f.write(text)
    else:
      logging.info(text)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.random_agent."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
import numpy as np
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import random_agent


class RandomAgentTest(absltest.TestCase):

  def test_step(self):
    agent = random_agent.RandomAgent(player_id=0, num_actions=10)

    legal_actions = [0, 2, 3, 5]
    time_step = rl_environment.TimeStep(
        observations={
            "info_state": [[0], [1]],
            "legal_actions": [legal_actions, []],
            "current_player": 0
        },
        rewards=None,
        discounts=None,
        step_type=None)
    agent_output = agent.step(time_step)

    self.assertIn(agent_output.action, legal_actions)
    self.assertAlmostEqual(sum(agent_output.probs), 1.0)
    self.assertEqual(
        len([x for x in agent_output.probs if x > 0]), len(legal_actions))
    self.assertTrue(
        np.allclose(agent_output.probs[legal_actions], [.25] * 4, atol=1e-5))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for open_spiel.python.algorithms.playthrough."""

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms import generate_playthrough


class PlaythroughTest(absltest.TestCase):

  def test_runs(self):
    result = generate_playthrough.playthrough(
        "tic_tac_toe", action_sequence=[0, 1, 2, 3, 4, 5, 6, 7, 8])
    self.assertNotEmpty(result)

  def test_format_tensor_1d(self):
    lines = generate_playthrough._format_tensor(np.array((1, 0, 1, 1)), "x")
    self.assertEqual(lines, ["x: ◉◯◉◉"])

  def test_format_tensor_2d(self):
    lines = generate_playthrough._format_tensor(np.array(((1, 0), (1, 1))), "x")
    self.assertEqual(lines, [
        "x: ◉◯",
        "   ◉◉",
    ])

  def test_format_tensor_3d(self):
    lines = []
    tensor = np.array((
        ((1, 0), (1, 1)),
        ((0, 0), (1, 0)),
        ((0, 1), (1, 0)),
    ))
    lines = generate_playthrough._format_tensor(tensor, "x")
    self.assertEqual(lines, [
        "x:",
        "◉◯  ◯◯  ◯◉",
        "◉◉  ◉◯  ◉◯",
    ])

  def test_format_tensor_3d_linewrap(self):
    tensor = np.array((
        ((1, 0), (1, 1)),
        ((0, 0), (1, 0)),
        ((0, 1), (1, 0)),
    ))
    lines = generate_playthrough._format_tensor(tensor, "x", max_cols=9)
    self.assertEqual(lines, [
        "x:",
        "◉◯  ◯◯",
        "◉◉  ◉◯",
        "",
        "◯◉",
        "◉◯",
    ])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tabular Q-Learner self-play example.

Two Q-Learning agents are trained by playing against each other.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
from absl import app
from absl import flags
import numpy as np
from six.moves import range

from open_spiel.python import rl_environment
from open_spiel.python import rl_tools
from open_spiel.python.algorithms import tabular_qlearner

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_train_episodes", int(1e6),
                     "Number of training episodes.")
flags.DEFINE_integer("num_eval_episodes", int(1e4),
                     "Number of episodes to use during each evaluation.")
flags.DEFINE_integer("eval_freq", int(1e4),
                     "The frequency (in episodes) to run evaluation.")
flags.DEFINE_string(
    "epsilon_schedule", None,
    "Epsilon schedule: e.g. 'linear,init,final,num_steps' or "
    "'constant,0.2'")
flags.DEFINE_string("game", "tic_tac_toe", "Game to load.")


def eval_agents(env, agents, num_episodes):
  """Evaluate the agents, returning a numpy array of average returns."""
  rewards = np.array([0] * env.num_players, dtype=np.float64)
  for _ in range(num_episodes):
    time_step = env.reset()
    while not time_step.last():
      player_id = time_step.observations["current_player"]
      agent_output = agents[player_id].step(time_step, is_evaluation=True)
      time_step = env.step([agent_output.action])
    for i in range(env.num_players):
      rewards[i] += time_step.rewards[i]
  rewards /= num_episodes
  return rewards


def create_epsilon_schedule(sched_str):
  """Creates an epsilon schedule from the string as desribed in the flags."""
  values = FLAGS.epsilon_schedule.split(",")
  if values[0] == "linear":
    assert len(values) == 4
    return rl_tools.LinearSchedule(
        float(values[1]), float(values[2]), int(values[3]))
  elif values[0] == "constant":
    assert len(values) == 2
    return rl_tools.ConstantSchedule(float(values[1]))
  else:
    print("Unrecognized schedule string: {}".format(sched_str))
    sys.exit()


def main(_):
  env = rl_environment.Environment(FLAGS.game)
  num_players = env.num_players
  num_actions = env.action_spec()["num_actions"]

  agents = []
  if FLAGS.epsilon_schedule is not None:
    for idx in range(num_players):
      agents.append(
          tabular_qlearner.QLearner(
              player_id=idx,
              num_actions=num_actions,
              epsilon_schedule=create_epsilon_schedule(FLAGS.epsilon_schedule)))
  else:
    agents = [
        tabular_qlearner.QLearner(player_id=idx, num_actions=num_actions)
        for idx in range(num_players)
    ]

  # 1. Train the agents
  training_episodes = FLAGS.num_train_episodes
  for cur_episode in range(training_episodes):
    if cur_episode % int(FLAGS.eval_freq) == 0:
      avg_rewards = eval_agents(env, agents, FLAGS.num_eval_episodes)
      print("Training episodes: {}, Avg rewards: {}".format(
          cur_episode, avg_rewards))
    time_step = env.reset()
    while not time_step.last():
      player_id = time_step.observations["current_player"]
      agent_output = agents[player_id].step(time_step)
      time_step = env.step([agent_output.action])

    # Episode is over, step all agents with final info state.
    for agent in agents:
      agent.step(time_step)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit test for GamutGenerator."""

from absl import app
from absl.testing import absltest
import pyspiel


class GamutGeneratorTest(absltest.TestCase):

  def test_generate_game(self):
    generator = pyspiel.GamutGenerator(
        "gamut.jar")

    # Using a list of arguments.
    game = generator.generate_game([
        "-g", "RandomGame", "-players", "4", "-normalize", "-min_payoff", "0",
        "-max_payoff", "150", "-actions", "2", "4", "5", "7"
    ])
    self.assertIsNotNone(game)

    # Using a string of arguments.
    game = generator.generate_game(
        "-g RandomGame -players 4 -normalize -min_payoff 0 " +
        "-max_payoff 150 -actions 2 4 5 7")
    self.assertIsNotNone(game)


def main(_):
  absltest.main()


if __name__ == "__main__":
  # Calling main via app.run here is necessary for internal uses.
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Computes the value of a given policy."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

PROBABILITY_THRESHOLD = 0


def policy_value(state, policies):
  """Returns the expected values for the state for players following `policies`.

  Computes the expected value of the`state` for each player, assuming player `i`
  follows the policy given in `policies[i]`.

  Args:
    state: A `pyspiel.State`.
    policies: A `list` of `policy.Policy` objects, one per player.

  Returns:
    A `numpy.array` containing the expected value for each player.
  """
  assert not state.is_simultaneous_node()

  num_players = len(policies)

  if state.is_terminal():
    values = np.array(state.returns())
  elif state.is_chance_node():
    values = np.zeros(shape=num_players)
    for action, action_prob in state.chance_outcomes():
      child = state.child(action)
      values += action_prob * policy_value(child, policies)
    return values
  elif state.is_simultaneous_node():
    raise NotImplementedError(
        "Policy Value is not implemented for simultaneous games")
  else:
    player = state.current_player()
    values = np.zeros(shape=num_players)
    for action, probability in policies[player].action_probabilities(
        state).items():
      if probability > PROBABILITY_THRESHOLD:
        child = state.child(action)
        values += probability * policy_value(child, policies)
  return values



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Q-values and reach probabilities computation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import numpy as np

_CalculatorReturn = collections.namedtuple(
    "_CalculatorReturn",
    [
        # A list of size `num_players` of the root node value for each player.
        "root_node_values",
        # An array of shape `[len(info_states), game.num_distinct_actions()]`
        # giving the value of each action. Will be zero for invalid actions.
        "action_values",
        # The player's counterfactual reach probability of this infostate when
        # playing, as a list of shape [num_info_states].
        "counterfactual_reach_probs",
        # The reach probability of the current player at the infostates, as a
        # list of shape [num_info_states].
        # This is the product of the current player probs along *one* trajectory
        # leading to this info-state (this number should be the same along
        # any trajectory leading to this info-state because of perfect recall).
        "player_reach_probs",
        # A list of `len(info_states)` `[game.num_distinct_actions()]` numpy
        # array so that v[s_index][a] = \sum_{h \in x} cfr_reach(h) * Q(h, a)
        "sum_cfr_reach_by_action_value",
    ])


class TreeWalkCalculator(object):
  r"""Class to orchestrate the calculation.

  This performs a full history tree walk and computes several statistics,
  available as attributes.

  Attributes:
    weighted_action_values: A dictionary mapping (player,information state
      string) to a dictionary mapping each action to a vector of the sum of
      (reward * prob) reward taking that action for each player. To get the
      action-values, one will need to normalize by `info_state_prob`.
    info_state_prob:  A dictionary mapping (player,information state string) to
      the reach probability of this info_state.
    info_state_player_prob: Same as info_state_prob for the player reach
      probability.
    info_state_cf_prob: Same as info_state_prob for the counterfactual reach
      probability to get to that state, i.e. the sum over histories, of the
      product of the opponents probabilities of actions leading to the history.
    info_state_chance_prob: Same as above, for the chance probability to get
      into that state.
    info_state_cf_prob_by_q_sum: A dictionary mapping (player,information state
      string) to a vector of shape `[num_actions]`, that store for each action
      the cumulative \sum_{h \in x} cfr_reach(h) * Q(h, a)
    root_values: The values at the root node [for player 0, for player 1].
  """

  def __init__(self, game):
    if not game.get_type().provides_information_state_string:
      raise ValueError("Only game which provide the information_state_string "
                       "are supported, as this is being used in the key to "
                       "identify states.")

    self._game = game
    self._num_players = game.num_players()
    self._num_actions = game.num_distinct_actions()

    self.weighted_action_values = None
    self.info_state_prob = None
    self.info_state_player_prob = None
    self.info_state_cf_prob = None
    self.info_state_chance_prob = None
    self.info_state_cf_prob_by_q_sum = None
    self.root_values = None

  def _get_action_values(self, state, policies, reach_probabilities):
    """Computes the value of the state given the policies for both players.

    Args:
      state: The state to start analysis from.
      policies: List of `policy.Policy` objects, one per player.
      reach_probabilities: A numpy array of shape `[num_players + 1]`.
        reach_probabilities[i] is the product of the player i action
        probabilities along the current trajectory. Note that
        reach_probabilities[-1] corresponds to the chance player. Initially, it
        should be called with np.ones(self._num_players + 1) at the root node.

    Returns:
      The value of the root state to each player.

    Side-effects - populates:
      `self.weighted_action_values[(player, infostate)][action]`.
      `self.info_state_prob[(player, infostate)]`.
      `self.info_state_cf_prob[(player, infostate)]`.
      `self.info_state_chance_prob[(player, infostate)]`.

    We use `(player, infostate)` as a key in case the same infostate is shared
    by multiple players, e.g. in a simultaneous-move game.
    """
    if state.is_terminal():
      return np.array(state.returns())

    current_player = state.current_player()
    is_chance = state.is_chance_node()

    if not is_chance:
      key = (current_player, state.information_state_string())
      reach_prob = np.prod(reach_probabilities)

      # We exclude both the current and the chance players.
      opponent_probability = (
          np.prod(reach_probabilities[:current_player]) *
          np.prod(reach_probabilities[current_player + 1:-1]))
      self.info_state_cf_prob[key] += (
          reach_probabilities[-1] * opponent_probability)
      self.info_state_prob[key] += reach_prob
      self.info_state_chance_prob[key] += reach_probabilities[-1]
      # Mind that we have "=" here and not "+=", because we just need to use
      # the reach prob for the player for *any* of the histories leading to
      # the current info_state (they are all equal because of perfect recall).
      self.info_state_player_prob[key] = reach_probabilities[current_player]

    value = np.zeros(len(policies))
    if is_chance:
      action_to_prob = dict(state.chance_outcomes())
    else:
      action_to_prob = policies[current_player].action_probabilities(state)
    for action in state.legal_actions():
      prob = action_to_prob.get(action, 0)
      new_reach_probabilities = reach_probabilities.copy()
      new_reach_probabilities[current_player] *= prob

      child = state.child(action)
      child_value = self._get_action_values(
          child, policies, reach_probabilities=new_reach_probabilities)
      if not is_chance:
        self.weighted_action_values[key][action] += child_value * reach_prob
        self.info_state_cf_prob_by_q_sum[key][action] += (
            child_value[current_player] * opponent_probability *
            reach_probabilities[-1])
      value += child_value * prob
    return value

  def compute_all_states_action_values(self, policies):
    """Computes action values per state for the player.

    The internal state is fully re-created when calling this method, thus it's
    safe to use one object to perform several tree-walks using different
    policies, and to extract the results using for example
    `calculator.infor_state_prob` to take ownership of the dictionary.

    Args:
      policies: List of `policy.Policy` objects, one per player. As the policy
        will be accessed using `policies[i]`, it can also be a dictionary
        mapping player_id to a `policy.Policy` object.
    """
    assert len(policies) == self._num_players

    # Compute action values
    self.weighted_action_values = collections.defaultdict(
        lambda: collections.defaultdict(lambda: np.zeros(self._num_players)))
    self.info_state_prob = collections.defaultdict(float)
    self.info_state_player_prob = collections.defaultdict(float)
    self.info_state_cf_prob = collections.defaultdict(float)
    self.info_state_chance_prob = collections.defaultdict(float)
    self.info_state_cf_prob_by_q_sum = collections.defaultdict(
        lambda: np.zeros(self._num_actions))

    self.root_values = self._get_action_values(
        self._game.new_initial_state(),
        policies,
        reach_probabilities=np.ones(self._num_players + 1))

  def _get_tabular_statistics(self, keys):
    """Returns tabular numpy arrays of the resulting stastistics.

    Args:
      keys: A list of the (player, info_state_str) keys to use to return the
        tabular numpy array of results.
    """
    # Collect normalized action values for each information state
    action_values = []
    cfrp = []  # Counterfactual reach probabilities
    player_reach_probs = []
    sum_cfr_reach_by_action_value = []

    for key in keys:
      player = key[0]
      av = self.weighted_action_values[key]
      norm_prob = self.info_state_prob[key]
      action_values.append([(av[a][player] / norm_prob) if
                            (a in av and norm_prob > 0) else 0
                            for a in range(self._num_actions)])
      cfrp.append(self.info_state_cf_prob[key])
      player_reach_probs.append(self.info_state_player_prob[key])
      sum_cfr_reach_by_action_value.append(
          self.info_state_cf_prob_by_q_sum[key])

    # Return values
    return _CalculatorReturn(
        root_node_values=self.root_values,
        action_values=action_values,
        counterfactual_reach_probs=cfrp,
        player_reach_probs=player_reach_probs,
        sum_cfr_reach_by_action_value=sum_cfr_reach_by_action_value)

  def get_tabular_statistics(self, tabular_policy):
    """Returns tabular numpy arrays of the resulting stastistics.

    This function should be called after `compute_all_states_action_values`.
    Optionally, one can directly call the object to perform both actions.

    Args:
      tabular_policy: A `policy.TabularPolicy` object, used to get the ordering
        of the states in the tabular numpy array.
    """
    keys = []
    for player_id, player_states in enumerate(tabular_policy.states_per_player):
      keys += [(player_id, s) for s in player_states]
    return self._get_tabular_statistics(keys)

  def __call__(self, policies, tabular_policy):
    """Computes action values per state for the player.

    The internal state is fully re-created when calling this method, thus it's
    safe to use one object to perform several tree-walks using different
    policies, and to extract the results using for example
    `calculator.infor_state_prob` to take ownership of the dictionary.

    Args:
      policies: List of `policy.Policy` objects, one per player.
      tabular_policy: A `policy.TabularPolicy` object, used to get the ordering
        of the states in the tabular numpy array.

    Returns:
      A `_CalculatorReturn` namedtuple. See its docstring for the details.
    """
    self.compute_all_states_action_values(policies)
    return self.get_tabular_statistics(tabular_policy)

  def get_root_node_values(self, policies):
    """Gets root values only.

    This speeds up calculation in two ways:

    1. It only searches nodes with positive probability.
    2. It does not populate a large dictionary of meta information.

    Args:
      policies: List of `policy.Policy` objects, one per player.

    Returns:
      A numpy array of shape [num_players] of the root value.
    """
    return self._get_action_values_only(
        self._game.new_initial_state(),
        policies,
        reach_probabilities=np.ones(self._num_players + 1))

  def _get_action_values_only(self, state, policies, reach_probabilities):
    """Computes the value of the state given the policies for both players.

    Args:
      state: The state to start analysis from.
      policies: List of `policy.Policy` objects, one per player.
      reach_probabilities: A numpy array of shape `[num_players + 1]`.
        reach_probabilities[i] is the product of the player i action
        probabilities along the current trajectory. Note that
        reach_probabilities[-1] corresponds to the chance player. Initially, it
        should be called with np.ones(self._num_players + 1) at the root node.

    Returns:
      A numpy array of shape [num_players] of the root value.
    """
    if state.is_terminal():
      return np.array(state.returns())

    current_player = state.current_player()
    is_chance = state.is_chance_node()

    value = np.zeros(len(policies))
    if is_chance:
      action_to_prob = dict(state.chance_outcomes())
    else:
      action_to_prob = policies[current_player].action_probabilities(state)

    for action in state.legal_actions():
      prob = action_to_prob.get(action, 0)

      # Do not follow tree down if there is zero probability.
      if prob == 0.0:
        continue

      new_reach_probabilities = reach_probabilities.copy()
      new_reach_probabilities[current_player] *= prob

      child = state.child(action)
      child_value = self._get_action_values_only(
          child, policies, reach_probabilities=new_reach_probabilities)
      value += child_value * prob
    return value



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN agents trained on Breakthrough by independent Q-learning."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
from absl import logging
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python import rl_environment
from open_spiel.python.algorithms import dqn
from open_spiel.python.algorithms import random_agent

FLAGS = flags.FLAGS

# Training parameters
flags.DEFINE_string("checkpoint_dir", "/tmp/dqn_test",
                    "Directory to save/load the agent models.")
flags.DEFINE_integer(
    "save_every", int(1e4),
    "Episode frequency at which the DQN agent models are saved.")
flags.DEFINE_integer("num_train_episodes", int(1e6),
                     "Number of training episodes.")
flags.DEFINE_integer(
    "eval_every", 1000,
    "Episode frequency at which the DQN agents are evaluated.")

# DQN model hyper-parameters
flags.DEFINE_list("hidden_layers_sizes", [64, 64],
                  "Number of hidden units in the Q-Network MLP.")
flags.DEFINE_integer("replay_buffer_capacity", int(1e5),
                     "Size of the replay buffer.")
flags.DEFINE_integer("batch_size", 32,
                     "Number of transitions to sample at each learning step.")


def eval_against_random_bots(env, trained_agents, random_agents, num_episodes):
  """Evaluates `trained_agents` against `random_agents` for `num_episodes`."""
  num_players = len(trained_agents)
  sum_episode_rewards = np.zeros(num_players)
  for player_pos in range(num_players):
    cur_agents = random_agents[:]
    cur_agents[player_pos] = trained_agents[player_pos]
    for _ in range(num_episodes):
      time_step = env.reset()
      episode_rewards = 0
      while not time_step.last():
        player_id = time_step.observations["current_player"]
        if env.is_turn_based:
          agent_output = cur_agents[player_id].step(
              time_step, is_evaluation=True)
          action_list = [agent_output.action]
        else:
          agents_output = [
              agent.step(time_step, is_evaluation=True) for agent in cur_agents
          ]
          action_list = [agent_output.action for agent_output in agents_output]
        time_step = env.step(action_list)
        episode_rewards += time_step.rewards[player_pos]
      sum_episode_rewards[player_pos] += episode_rewards
  return sum_episode_rewards / num_episodes


def main(_):
  game = "breakthrough"
  num_players = 2

  env_configs = {"columns": 5, "rows": 5}
  env = rl_environment.Environment(game, **env_configs)
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  # random agents for evaluation
  random_agents = [
      random_agent.RandomAgent(player_id=idx, num_actions=num_actions)
      for idx in range(num_players)
  ]

  with tf.Session() as sess:
    hidden_layers_sizes = [int(l) for l in FLAGS.hidden_layers_sizes]
    # pylint: disable=g-complex-comprehension
    agents = [
        dqn.DQN(
            session=sess,
            player_id=idx,
            state_representation_size=info_state_size,
            num_actions=num_actions,
            hidden_layers_sizes=hidden_layers_sizes,
            replay_buffer_capacity=FLAGS.replay_buffer_capacity,
            batch_size=FLAGS.batch_size) for idx in range(num_players)
    ]
    sess.run(tf.global_variables_initializer())

    for ep in range(FLAGS.num_train_episodes):
      if (ep + 1) % FLAGS.eval_every == 0:
        r_mean = eval_against_random_bots(env, agents, random_agents, 1000)
        logging.info("[%s] Mean episode rewards %s", ep + 1, r_mean)
      if (ep + 1) % FLAGS.save_every == 0:
        for agent in agents:
          agent.save(FLAGS.checkpoint_dir)

      time_step = env.reset()
      while not time_step.last():
        player_id = time_step.observations["current_player"]
        if env.is_turn_based:
          agent_output = agents[player_id].step(time_step)
          action_list = [agent_output.action]
        else:
          agents_output = [agent.step(time_step) for agent in agents]
          action_list = [agent_output.action for agent_output in agents_output]
        time_step = env.step(action_list)

      # Episode is over, step all agents with final info state.
      for agent in agents:
        agent.step(time_step)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for OpenSpiel observation histories."""

from absl.testing import absltest

import pyspiel


def state_from_sequence(game, action_sequence):
  state = game.new_initial_state()
  for action in action_sequence:
    state.apply_action(action)
  return state


class ObservationHistoryTest(absltest.TestCase):

  def test_invalid_aoh_construction_raises_error(self):
    with self.assertRaises(RuntimeError):
      pyspiel.ActionObservationHistory(0, ["not tuple"])
    with self.assertRaises(RuntimeError):
      pyspiel.ActionObservationHistory(0, [("tuple", "too", "long")])
    with self.assertRaises(RuntimeError):
      pyspiel.ActionObservationHistory(0, [("not int", "obs")])
    with self.assertRaises(RuntimeError):
      pyspiel.ActionObservationHistory(0, [(1, 1)])  # obs not string

  def test_kuhn_rollout(self):
    game = pyspiel.load_game("kuhn_poker")

    # Test on this specific sequence.
    action_sequence = [2, 1, 0, 1, 1]

    root_state = game.new_initial_state()
    terminal = state_from_sequence(game, action_sequence)
    self.assertTrue(terminal.is_terminal())

    # Check prefixes, extensions and correspondences of both
    # public-observation histories / action-observation histories
    # while rolling out.
    seq_idx = 0

    def advance_and_test(state):
      nonlocal seq_idx
      action = action_sequence[seq_idx]

      parent_state = state.clone()
      state.apply_action(action)

      # Both PublicObservationHistory and ActionObservationHistory need to pass
      # the same correspondence / prefix / extension tests, for both
      # construction from state (and player for AOH) or just passing state (and
      # player) without history construction. So this tests them together.
      # These are passed as constructors (lefts) and targets (rights).
      constructors_with_targets = [
          # (p)layer, (s)tate
          (pyspiel.ActionObservationHistory, [
              lambda p, s: [pyspiel.ActionObservationHistory(p, s)],
              lambda p, s: [p, s]
          ]),
          # Since PublicObservationHistory does not need a player it is ignored.
          (lambda _, s: pyspiel.PublicObservationHistory(s), [
              lambda _, s: [pyspiel.PublicObservationHistory(s)],
              lambda _, s: [s]
          ])
      ]

      for (left, rights) in constructors_with_targets:
        for right in rights:
          for player in range(2):
            # Shortcuts for conciseness: the most important things
            # relevant for the tests have long names.
            # In other words, when reading the code just skip over
            # the short vars to gain understanding.
            l, r, p = left, right, player

            self.assertTrue(l(p, state).corresponds_to(*r(p, state)))
            self.assertFalse(l(p, parent_state).corresponds_to(*r(p, state)))
            if state.is_terminal():
              self.assertTrue(l(p, terminal).corresponds_to(*r(p, state)))
            else:
              self.assertFalse(l(p, terminal).corresponds_to(*r(p, state)))
            if state.is_initial_state():
              self.assertTrue(l(p, root_state).corresponds_to(*r(p, state)))
            else:
              self.assertFalse(l(p, root_state).corresponds_to(*r(p, state)))

            self.assertTrue(l(p, parent_state).is_prefix_of(*r(p, state)))
            self.assertFalse(l(p, state).is_prefix_of(*r(p, parent_state)))
            self.assertTrue(l(p, root_state).is_prefix_of(*r(p, state)))
            if state.is_terminal():
              self.assertTrue(l(p, terminal).is_prefix_of(*r(p, state)))
            else:
              self.assertFalse(l(p, terminal).is_prefix_of(*r(p, state)))

            self.assertFalse(l(p, parent_state).is_extension_of(*r(p, state)))
            self.assertTrue(l(p, state).is_extension_of(*r(p, parent_state)))
            self.assertTrue(l(p, terminal).is_extension_of(*r(p, state)))
            if state.is_initial_state():
              self.assertTrue(l(p, root_state).is_extension_of(*r(p, state)))
            else:
              self.assertFalse(l(p, root_state).is_extension_of(*r(p, state)))

      seq_idx += 1
      return state

    state = game.new_initial_state()
    self.assertTrue(state.is_chance_node())
    self.assertEqual(
        pyspiel.PublicObservationHistory(state),
        pyspiel.PublicObservationHistory([pyspiel.PublicObservation.START_GAME
                                         ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(0, state),
        pyspiel.ActionObservationHistory(
            0, [(None, pyspiel.PrivateObservation.NOTHING)]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(1, state),
        pyspiel.ActionObservationHistory(
            1, [(None, pyspiel.PrivateObservation.NOTHING)]))

    advance_and_test(state)
    self.assertTrue(state.is_chance_node())
    self.assertEqual(
        pyspiel.PublicObservationHistory(state),
        pyspiel.PublicObservationHistory(
            [pyspiel.PublicObservation.START_GAME, "Deal to player 0"]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(0, state),
        pyspiel.ActionObservationHistory(0, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "211")
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(1, state),
        pyspiel.ActionObservationHistory(1, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, pyspiel.PrivateObservation.NOTHING)
        ]))

    advance_and_test(state)
    self.assertTrue(state.is_player_node())
    self.assertEqual(
        pyspiel.PublicObservationHistory(state),
        pyspiel.PublicObservationHistory([
            pyspiel.PublicObservation.START_GAME, "Deal to player 0",
            "Deal to player 1"
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(0, state),
        pyspiel.ActionObservationHistory(0, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "211"),
            (None, "211")
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(1, state),
        pyspiel.ActionObservationHistory(1, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "111")
        ]))

    advance_and_test(state)
    self.assertTrue(state.is_player_node())
    self.assertEqual(
        pyspiel.PublicObservationHistory(state),
        pyspiel.PublicObservationHistory([
            pyspiel.PublicObservation.START_GAME, "Deal to player 0",
            "Deal to player 1", "Pass"
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(0, state),
        pyspiel.ActionObservationHistory(0, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "211"),
            (None, "211"),
            (0, "211")
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(1, state),
        pyspiel.ActionObservationHistory(1, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "111"),
            (None, "111")
        ]))

    advance_and_test(state)
    self.assertTrue(state.is_player_node())
    self.assertEqual(
        pyspiel.PublicObservationHistory(state),
        pyspiel.PublicObservationHistory([
            pyspiel.PublicObservation.START_GAME, "Deal to player 0",
            "Deal to player 1", "Pass", "Bet"
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(0, state),
        pyspiel.ActionObservationHistory(0, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "211"),
            (None, "211"),
            (0, "211"),
            (None, "212")
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(1, state),
        pyspiel.ActionObservationHistory(1, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "111"),
            (None, "111"),
            (1, "112")
        ]))

    advance_and_test(state)
    self.assertTrue(state.is_terminal())
    self.assertEqual(
        pyspiel.PublicObservationHistory(state),
        pyspiel.PublicObservationHistory([
            pyspiel.PublicObservation.START_GAME, "Deal to player 0",
            "Deal to player 1", "Pass", "Bet", "Bet"
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(0, state),
        pyspiel.ActionObservationHistory(0, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "211"),
            (None, "211"),
            (0, "211"),
            (None, "212"),
            (1, "222")
        ]))
    self.assertEqual(
        pyspiel.ActionObservationHistory(1, state),
        pyspiel.ActionObservationHistory(1, [
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, pyspiel.PrivateObservation.NOTHING),
            (None, "111"),
            (None, "111"),
            (1, "112"),
            (None, "122")
        ]))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Public tree visualization example.

Example usage:

  python public_tree_viz_example.py --game="kuhn_poker"  \
      --target_pubset="start game,Deal to player 0,Deal to player 1,Bet"
"""

from absl import app
from absl import flags
from absl import logging

import pyspiel
from open_spiel.python.visualizations import public_tree_viz
from open_spiel.python.visualizations import treeviz

FLAGS = flags.FLAGS
flags.DEFINE_string("game", "kuhn_poker", "Name of the game")
flags.DEFINE_string("out", "/tmp/gametree.png", "Name of output file, e.g., "
                    "[*.png|*.pdf].")
flags.DEFINE_enum("prog", "dot", ["dot", "neato", "circo"], "Graphviz layout.")
flags.DEFINE_string(
    "target_pubset", None,
    "Limit grouping of public states only to the specified "
    "public state.")
flags.DEFINE_boolean("verbose", False, "Whether to print verbose output.")


def _zero_sum_node_decorator(state):
  """Custom node decorator that only shows the return of the first player."""
  attrs = treeviz.default_node_decorator(state)  # get default attributes
  if state.is_terminal():
    attrs["label"] = str(int(state.returns()[0]))
  return attrs


def main(argv):
  del argv

  game = pyspiel.load_game_with_public_states(FLAGS.game)
  base_game = game.get_base_game()
  game_type = base_game.get_type()
  two_pl_zero_sum = game_type.utility == pyspiel.GameType.Utility.ZERO_SUM \
    and base_game.num_players() == 2
  node_decorator = _zero_sum_node_decorator if two_pl_zero_sum \
    else treeviz.default_node_decorator

  gametree = public_tree_viz.GamePublicTree(
      public_game=game,
      draw_world=True,
      target_public_to_base=FLAGS.target_pubset,
      node_decorator=node_decorator,
      group_pubsets=True)

  if FLAGS.verbose:
    logging.info("Game tree:\n%s", gametree.to_string())

  gametree.draw(FLAGS.out, prog=FLAGS.prog)
  logging.info("Game tree saved to file: %s", FLAGS.out)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Policy aggregator.

Turns a weighted sum of N policies into a realization-equivalent single
policy by sweeping over the state space.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from open_spiel.python import policy
import pyspiel


class PolicyFunction(policy.Policy):
  """A callable policy class."""

  def __init__(self, pids, policies, game):
    """Construct a policy function.

    Arguments:
      pids: spiel player id of players these policies belong to.
      policies: a list of dictionaries of keys (stringified binary observations)
        to a list of probabilities for each move uid (between 0 and max_moves -
        1).
      game: OpenSpiel game.
    """
    self._pids = pids
    self._policies = policies
    self._game = game
    self._game_type = game.get_type()

  def _state_key(self, state, player_id=None):
    """Returns the key to use to look up this (state, player_id) pair."""
    if self._game_type.provides_information_state_string:
      if player_id is None:
        return state.information_state_string()
      else:
        return state.information_state_string(player_id)
    elif self._game_type.provides_observation_tensor:
      if player_id is None:
        return state.observation_tensor()
      else:
        return state.observation_tensor(player_id)
    else:
      return str(state)

  @property
  def policy(self):
    return self._policies

  def action_probabilities(self, state, player_id=None):
    """Returns the policy for a player in a state.

    Args:
      state: A `pyspiel.State` object.
      player_id: Optional, the player id for whom we want an action. Optional
        unless this is a simultaneous state at which multiple players can act.

    Returns:
      A `dict` of `{action: probability}` for the specified player in the
      supplied state.
    """
    state_key = self._state_key(state, player_id=player_id)
    if state.is_simultaneous_node():
      # Policy aggregator doesn't yet support simultaneous moves nodes.
      # The below lines are one step towards that direction.
      result = []
      for player_pol in self._policies:
        result.append(player_pol[state_key])
      return result
    if player_id is None:
      player_id = state.current_player()
    return self._policies[player_id][state_key]


class PolicyPool(object):
  """Transforms a list of list of policies (One list per player) to callable."""

  def __init__(self, policies):
    """Transforms a list of list of policies (One list per player) to callable.

    Args:
      policies: List of list of policies.
    """
    self._policies = policies

  def __call__(self, state, player):
    return [
        a.action_probabilities(state, player_id=player)
        for a in self._policies[player]
    ]


class PolicyAggregator(object):
  """Main aggregator object."""

  def __init__(self, game, epsilon=1e-40):
    self._game = game
    self._game_type = game.get_type()
    self._num_players = self._game.num_players()
    self._policy_pool = None
    self._weights = None
    self._policy = {}
    self._epsilon = epsilon

  def _state_key(self, state, player_id=None):
    """Returns the key to use to look up this (state, player) pair."""
    # TODO(somidshafiei): fuse this with the identical PolicyFunction._state_key
    if self._game_type.provides_information_state_string:
      if player_id is None:
        return state.information_state_string()
      else:
        return state.information_state_string(player_id)
    elif self._game_type.provides_observation_string:
      if player_id is None:
        return state.observation_string()
      else:
        return state.observation_string(player_id)
    else:
      return str(state)

  def aggregate(self, pids, policies, weights):
    """Aggregate the list of policies for each player.

    Arguments:
      pids: the spiel player ids of the players the strategies belong to.
      policies: List of list of policies (One list per player)
      weights: the list of weights to attach to each policy.

    Returns:
      A PolicyFunction, a callable object representing the policy.
    """
    aggr_policies = []

    for pid in pids:
      aggr_policies.append(self._sub_aggregate(pid, policies, weights))
    return PolicyFunction(pids, aggr_policies, self._game)

  def _sub_aggregate(self, pid, policies, weights):
    """Aggregate the list of policies for one player.

    Arguments:
      pid: the spiel player id of the player the strategies belong to.
      policies: List of list of policies (One list per player)
      weights: the list of weights to attach to each policy.

    Returns:
      A PolicyFunction, a callable object representing the policy.
    """
    self._policy_pool = PolicyPool(policies)
    # ipdb.set_trace()

    assert self._policy_pool is not None
    self._weights = weights
    # string of state -> probs list
    self._policy = {}

    state = self._game.new_initial_state()
    my_reaches = weights[:]
    self._rec_aggregate(pid, state, my_reaches)

    # Now normalize
    for key in self._policy:
      actions, probabilities = zip(*self._policy[key].items())
      # Add some small proba mass to avoid divide by zero, which happens for
      # games with low reach probabilities for certain states (keys)
      new_probs = [prob + self._epsilon for prob in probabilities]
      denom = sum(new_probs)
      for i in range(len(actions)):
        self._policy[key][actions[i]] = new_probs[i] / denom
    return self._policy

  def _rec_aggregate(self, pid, state, my_reaches):
    """Recursively traverse game tree to compute aggregate policy."""

    if state.is_terminal():
      return
    elif state.is_simultaneous_node():
      # TODO(author10): this is assuming that if there is a sim.-move state, it is
      #               the only state, i.e., the game is a normal-form game
      assert self._game_type.dynamics ==\
        pyspiel.GameType.Dynamics.SIMULTANEOUS, "Game must be simultaneous-move"
      assert self._game_type.chance_mode ==\
        pyspiel.GameType.ChanceMode.DETERMINISTIC, "Chance nodes not supported"
      assert self._game_type.information ==\
        pyspiel.GameType.Information.ONE_SHOT, "Only one-shot NFGs supported"
      policies = self._policy_pool(state, pid)
      state_key = self._state_key(state, pid)
      self._policy[state_key] = {}
      for player_policy, weight in zip(policies, my_reaches[pid]):
        for action in player_policy.keys():
          if action in self._policy[state_key]:
            self._policy[state_key][action] += weight * player_policy[action]
          else:
            self._policy[state_key][action] = weight * player_policy[action]
      return
    elif state.is_chance_node():
      # do not factor in opponent reaches
      outcomes, _ = zip(*state.chance_outcomes())
      for i in range(0, len(outcomes)):
        outcome = outcomes[i]
        new_state = state.clone()
        new_state.apply_action(outcome)
        self._rec_aggregate(pid, new_state, my_reaches)
      return
    else:
      turn_player = state.current_player()

      state_key = self._state_key(state, turn_player)
      legal_policies = self._policy_pool(state, turn_player)
      if pid == turn_player:
        # update the current node
        # will need the observation to query the policies
        if state not in self._policy:
          self._policy[state_key] = {}

      used_moves = []
      for k in range(len(legal_policies)):
        used_moves += [a[0] for a in legal_policies[k].items()]
      used_moves = np.unique(used_moves)

      for uid in used_moves:
        new_reaches = np.copy(my_reaches)
        if pid == turn_player:
          for i in range(len(legal_policies)):
            # compute the new reach for each policy for this action
            new_reaches[turn_player][i] *= legal_policies[i].get(uid, 0)
            # add reach * prob(a) for this policy to the computed policy
            if uid in self._policy[state_key].keys():
              self._policy[state_key][uid] += new_reaches[turn_player][i]
            else:
              self._policy[state_key][uid] = new_reaches[turn_player][i]

        # recurse
        new_state = state.clone()
        new_state.apply_action(uid)
        self._rec_aggregate(pid, new_state, new_reaches)
